import numpy as np

import _example


class Example(_example.Example):
    title = 'Hello Program'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                in vec2 in_vert;

                void main() {
                    gl_Position = vec4(in_vert, 0.0, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330

                out vec4 f_color;

                void main() {
                    f_color = vec4(0.2, 0.4, 0.7, 1.0);
                }
            ''',
        )

        vertices = np.array([
            0.0, 0.8,
            -0.6, -0.8,
            0.6, -0.8,
        ])

        self.vbo = self.ctx.buffer(vertices.astype('f4').tobytes())
        self.vao = self.ctx.vertex_array(self.prog, self.vbo, 'in_vert')

    def render(self, time: float, frame_time: float):
        self.ctx.screen.clear(color=(1.0, 1.0, 1.0))
        self.vao.render()


if __name__ == '__main__':
    Example.run()



"""
A Game of Life implementation using transform feedback.

We calculate the next state of the map with transform()
meaning a vertex shader will generate the new state into a buffer.
This buffer is then written into the texture we display.
This is a fast vram to vram copy.

Comments:

Another way to do this is simply rendering to Framebuffers.
"""
import numpy as np

import moderngl
from ported._example import Example


class Conway(Example):
    title = "Conway's Game of Life"
    window_size = 800, 800

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # How often the map should be updated
        self.update_delay = 1 / 60  # updates per second
        self.last_updated = 0
        # size of the map
        self.width, self.height = 400, 400
        # Force the window to calculate black borders if needed to retain the aspect ratio
        self.wnd.fixed_aspect_ratio = self.width / self.height
        # Initial state of the map (random)
        pixels = np.round(np.random.rand(self.width, self.height)).astype('f4')

        # Program drawing the result to the screen.
        # This is rendered simply using a textured screen aligned triangle strip
        self.display_prog = self.ctx.program(
            vertex_shader='''
                #version 330

                in vec2 in_vert;
                in vec2 in_texcoord;

                out vec2 v_text;

                void main() {
                    v_text = in_texcoord;
                    gl_Position = vec4(in_vert, 0.0, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330

                // Will read from texture bound to channel / locaton 0 by default
                uniform sampler2D Texture;

                // Interpolated texture coordinate from vertex shader
                in vec2 v_text;
                // The fragment ending up on the screen
                out vec4 f_color;

                void main() {
                    f_color = texture(Texture, v_text);
                }
            ''',
        )

        # Program calculating the next state of the map
        self.transform_prog = self.ctx.program(
            vertex_shader='''
                #version 330

                uniform sampler2D Texture;

                out float out_vert;

                #define LIVING 0.0
                #define DEAD 1.0

                bool cell(int x, int y) {
                    // get the texture size
                    ivec2 tSize = textureSize(Texture, 0).xy;
                    // Ensure lookups are not going outside the texture area because
                    // texelFetch do not support texture repeat / wrap modes
                    return texelFetch(Texture, ivec2((x + tSize.x) % tSize.x, (y + tSize.y) % tSize.y), 0).r < 0.5;
                }

                void main() {
                    int width = textureSize(Texture, 0).x;
                    ivec2 in_text = ivec2(gl_VertexID % width, gl_VertexID / width);
                    bool living = cell(in_text.x, in_text.y);

                    int neighbours = 0;
                    if (cell(in_text.x - 1, in_text.y - 1)) neighbours++;
                    if (cell(in_text.x - 1, in_text.y + 0)) neighbours++;
                    if (cell(in_text.x - 1, in_text.y + 1)) neighbours++;
                    if (cell(in_text.x + 1, in_text.y - 1)) neighbours++;
                    if (cell(in_text.x + 1, in_text.y + 0)) neighbours++;
                    if (cell(in_text.x + 1, in_text.y + 1)) neighbours++;
                    if (cell(in_text.x + 0, in_text.y + 1)) neighbours++;
                    if (cell(in_text.x + 0, in_text.y - 1)) neighbours++;

                    if (living) {
                        out_vert = (neighbours == 2 || neighbours == 3) ? LIVING : DEAD;
                    } else {
                        out_vert = (neighbours == 3) ? LIVING : DEAD;
                    }
                }
            ''',
            varyings=['out_vert']
        )

        # Create the map texture
        self.texture = self.ctx.texture((self.width, self.height), 1, pixels.tobytes(), dtype='f4')
        self.texture.filter = moderngl.NEAREST, moderngl.NEAREST
        self.texture.swizzle = 'RRR1'  # What components texelFetch will get from the texture (in shader)

        # A quad covering the screen with texture coordinates
        self.vbo = self.ctx.buffer(np.array([
            # x    y     u  v
            -1.0, -1.0,  0, 0,  # lower left
            -1.0,  1.0,  0, 1,  # upper left
            1.0,  -1.0,  1, 0,  # lower right
            1.0,   1.0,  1, 1,  # upper right
        ], dtype="f4"))
        self.vao = self.ctx.simple_vertex_array(self.display_prog, self.vbo, 'in_vert', 'in_texcoord')

        # Transform vertex array to generate new map state
        self.tao = self.ctx.vertex_array(self.transform_prog, [])
        self.pbo = self.ctx.buffer(reserve=pixels.nbytes)  # buffer to store the result

    def render(self, time, frame_time):
        self.ctx.clear(1.0, 1.0, 1.0)

        # Bind texture to channel 0
        self.texture.use(location=0)

        if time - self.last_updated > self.update_delay:
            # Generate the new map and write that to the pbo buffer
            self.tao.transform(self.pbo, vertices=self.width * self.height)
            # Copy the pbo into the texture
            self.texture.write(self.pbo)
            self.last_updated = time

        # Render the texture
        self.vao.render(moderngl.TRIANGLE_STRIP)


if __name__ == '__main__':
    Conway.run()



"""
Example showing how to resize Buffers with orphan()

This can be useful for batch drawing an arbitrary
amount of geometry over time.

We just render a growing amount of random points.
"""
import struct
import random
import moderngl
from pyrr import matrix44
from ported._example import Example


class Points:
    """Simple point batching.

    The point set is created using an initial buffer allocation.
    When the buffer is to small we double the size.
    When the buffer do not need resizing we only do a partial
    buffer update.
    """
    def __init__(self, ctx, num_points):
        """
        Args:
            ctx: moderngl context
            num_points: Initial number of points to allocate
        """
        self.points = []
        self.ctx = ctx
        self.buffer = self.ctx.buffer(reserve=num_points * 12)  # 12 bytes for a 3f
        self.program = self.ctx.program(
            vertex_shader="""
            #version 330
            in vec3 in_position;
            uniform mat4 model_matrix;
            void main() {
                gl_Position = model_matrix * vec4(in_position, 1.0);
            }
            """,
            fragment_shader="""
            #version 330
            out vec4 outColor;
            void main() {
                outColor = vec4(1.0);
            }
            """,
        )
        self.vao = self.ctx.vertex_array(
            self.program,
            [(self.buffer, '3f', 'in_position')],
        )

    def render(self, time):
        self.program['model_matrix'].write(matrix44.create_from_eulers((0, time / 8, 0), dtype='f4'))
        self.vao.render(vertices=self.count, mode=moderngl.POINTS)

    @property
    def count(self):
        return len(self.points) // 3

    @property
    def byte_size(self):
        """int: Byte size of the point data"""
        return len(self.points) * 4  # 4 byte per float

    def add(self, num):
        """Adds num points random points"""
        resized = False
        old_points_size = self.byte_size
        new = list(self._gen_random_points(num))
        self.points = self.points + new

        # Keep doubling the buffer size until we reach an acceptable size
        while self.byte_size > self.buffer.size:
            resized = True
            print("Buffer resized {} -> {}".format(self.buffer.size, self.buffer.size * 2))
            print("New capacity is {} points".format(self.buffer.size * 2 // 12))
            self.buffer.orphan(self.buffer.size * 2)

        if resized:
            # Re-write the entire buffer
            self.buffer.write(struct.pack('{}f'.format(len(self.points)), *self.points))
        else:
            # Partial buffer update
            print("Partial buffer update adding {} points".format(len(new) // 3))
            self.buffer.write(struct.pack('{}f'.format(len(new)), *new), offset=old_points_size)

    def _gen_random_points(self, num):
        for _ in range(num * 3):
            yield random.uniform(-1.5, 1.5)


class GrowingBuffers(Example):
    """Growing buffers"""
    gl_version = (3, 3)
    title = "Buffer Resize / Batch Draw"
    window_size = 720, 720
    aspect_ratio = 1.0

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.points = Points(self.ctx, 12 * 10)
        self.points.add(10)

    def render(self, time, frametime):
        self.points.render(time)

        # Add more points every 60 frames
        if self.wnd.frames % 60 == 0:
            self.points.add(5000)


if __name__ == '__main__':
    GrowingBuffers.run()



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import sys
# import struct

# from OpenGL.GLUT import (
#     GLUT_DEPTH, GLUT_DOUBLE, GLUT_RGB,
#     glutCreateWindow, glutDisplayFunc, glutInit, glutInitDisplayMode,
#     glutInitWindowSize, glutMainLoop, glutSwapBuffers,
# )

# import ModernGL

# glutInit(sys.argv)
# glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH)
# glutInitWindowSize(800, 600)
# glutCreateWindow(b'')

# ctx = ModernGL.create_context()

# vert = ctx.vertex_shader('''
#     #version 330
#     in vec2 vert;
#     void main() {
#         gl_Position = vec4(vert, 0.0, 1.0);
#     }
# ''')

# frag = ctx.fragment_shader('''
#     #version 330
#     out vec4 color;
#     void main() {
#         color = vec4(0.30, 0.50, 1.00, 1.0);
#     }
# ''')

# prog = ctx.program(vert, frag])

# vbo = ctx.buffer(struct.pack('6f', 0.0, 0.8, -0.6, -0.8, 0.6, -0.8))
# vao = ctx.simple_vertex_array(prog, vbo, ['vert'])


# def display():
#     ctx.clear(0.9, 0.9, 0.9)
#     vao.render()
#     glutSwapBuffers()


# glutDisplayFunc(display)
# glutMainLoop()



"""
Emit particles from mouse positions with random velocities
adding the current velocity of the mouse pointer.

Shows how you can use transform feedback with custom emitting.
You can emit with the gpu or cpu. The main trick is to use
transform() with the buffer_offset parameter writing to be back
of the buffer. We use a query object in this example that can
potentially stall rendering a bit, but it's still orders of
magnitude faster than doing everything on the cpu. using
transform() with a buffer offset can be used in many creative ways.

This is similar to pausing and resuming transform feedbacks,
but works in GL 3.3 core.

* The geometry shader is used to destroy geometry for expired
  particles.
* We use a query to count how many primitives the geometry shader
  actually emitted so we can know how many new particles we should
  emit.

We show 3 different emit methods:
* Method #1: CPU emitting by manually writing new particles to the end of the buffer.
             This will probably stream up the entire buffer per frame, but is acceptable
             for smaller amount of data.
* Method #2: CPU emitting by filling a smaller emit buffer to reduce the amount of data we stream per frame
             Here we fill a smaller emit buffer and render that to the end of the output buffer
             to greatly reduce the amount of data we stream to the gpu per frame.
* Method #3: GPU emitting. A shader simply calculates the new values.
             No buffer streaming. All is on the gpu side. This method is
             orders of magnitude faster than method 1 and 2

Emit particles from mouse positions:

    mouse_control = True

Emit particles from a predetermined path:

    mouse_control = False

"""
import numpy as np

import moderngl
from moderngl_window import screenshot
from ported._example import Example
from pyrr import matrix44


class Particles(Example):
    title = "Particle System Emitter"
    gl_version = (3, 3)
    aspect_ratio = None
    mouse_control = False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Renders particles to the screen
        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                uniform mat4 projection;

                in vec2 in_pos;
                in vec3 in_color;
                out vec3 color;

                void main() {
                    gl_Position = projection * vec4(in_pos, 0.0, 1.0);
                    color = in_color;
                }
            ''',
            fragment_shader='''
                #version 330

                out vec4 f_color;
                in vec3 color;

                void main() {
                    f_color = vec4(color, 1.0);
                }
            ''',
        )

        # Animates / transforms the particles per frame
        self.transform = self.ctx.program(
            vertex_shader='''
            #version 330

            in vec2 in_pos;
            in vec2 in_vel;
            in vec3 in_color;

            out vec2 vs_vel;
            out vec3 vs_color;

            void main() {
                gl_Position = vec4(in_pos, 0.0, 1.0);
                vs_vel = in_vel;
                vs_color = in_color;
            }
            ''',
            geometry_shader='''
            #version 330

            layout(points) in;
            layout(points, max_vertices = 1) out;

            uniform float gravity;
            uniform float ft;

            in vec2 vs_vel[1];
            in vec3 vs_color[1];

            out vec2 out_pos;
            out vec2 out_vel;
            out vec3 out_color;

            void main() {
                vec2 pos = gl_in[0].gl_Position.xy;
                vec2 velocity = vs_vel[0];

                if (pos.y > -1.0) {
                    vec2 vel = velocity + vec2(0.0, gravity);
                    out_pos = pos + vel * ft;
                    out_vel = vel;
                    out_color = vs_color[0];
                    EmitVertex();
                    EndPrimitive();
                }
            }
            ''',
            varyings=['out_pos', 'out_vel', 'out_color'],
        )
        self.mouse_pos = 0, 0
        self.mouse_velocity = 0.001, 0.001  # default value must not be 0. any number is fine
        self.prog['projection'].write(self.projection)
        self.transform['gravity'].value = -.01  # affects the velocity of the particles over time
        self.ctx.point_size = self.wnd.pixel_ratio * 2  # point size

        self.N = 25_000  # particle count
        self.active_particles = self.N // 100  # Initial / current number of active particles
        self.max_emit_count = self.N // 100  # Maximum number of particles to emit per frame
        self.stride = 28  # byte stride for each vertex
        self.floats = 7
        # Note that passing dynamic=True probably doesn't mean anything to most drivers today
        self.vbo1 = self.ctx.buffer(reserve=self.N * self.stride)
        self.vbo2 = self.ctx.buffer(reserve=self.N * self.stride)
        # Write some initial particles
        self.vbo1.write(
            np.fromiter(
                self.gen_particles(self.active_particles),
                count=self.active_particles * self.floats,
                dtype='f4',
            )
        )

        # Transform vaos. We transform data back and forth to avoid buffer copy
        self.transform_vao1 = self.ctx.vertex_array(
            self.transform,
            [(self.vbo1, '2f 2f 3f', 'in_pos', 'in_vel', 'in_color')],
        )
        self.transform_vao2 = self.ctx.vertex_array(
            self.transform,
            [(self.vbo2, '2f 2f 3f', 'in_pos', 'in_vel', 'in_color')],
        )

        # Render vaos. The render to screen version of the transform vaos above
        self.render_vao1 = self.ctx.vertex_array(
            self.prog,
            [(self.vbo1, '2f 2x4 3f', 'in_pos', 'in_color')],
        )
        self.render_vao2 = self.ctx.vertex_array(
            self.prog,
            [(self.vbo2, '2f 2x4 3f', 'in_pos', 'in_color')],
        )

        # Setup for emit method #2. The emit buffer size is only max_emit_count.
        self.emit_buffer_elements = self.max_emit_count
        self.emit_buffer = self.ctx.buffer(reserve=self.emit_buffer_elements * self.stride)
        self.emit_buffer_prog = self.ctx.program(  # Siple shader just emitting a buffer
            vertex_shader='''
            # version 330
            in vec2 in_pos;
            in vec2 in_vel;
            in vec3 in_color;
            out vec2 out_pos;
            out vec2 out_vel;
            out vec3 out_color;
            void main() {
                out_pos = in_pos;
                out_vel = in_vel;
                out_color = in_color;
            }
            ''',
            varyings=['out_pos', 'out_vel', 'out_color'],
        )
        self.emit_buffer_vao = self.ctx.vertex_array(
            self.emit_buffer_prog,
            [(self.emit_buffer, '2f 2f 3f', 'in_pos', 'in_vel', 'in_color')],
        )

        # Setup for method #3: GPU emitting
        self.gpu_emitter_prog = self.ctx.program(
            vertex_shader='''
            # version 330
            #define M_PI 3.1415926535897932384626433832795
            uniform vec2 mouse_pos;
            uniform vec2 mouse_vel;
            uniform float time;

            out vec2 out_pos;
            out vec2 out_vel;
            out vec3 out_color;

            float rand(float n){return fract(sin(n) * 43758.5453123);}
            void main() {
                float a = mod(time * gl_VertexID, M_PI * 2);
                float r = clamp(rand(time + gl_VertexID), 0.1, 0.9);
                out_pos = mouse_pos;
                out_vel = vec2(sin(a), cos(a)) * r + mouse_vel;
                out_color = vec3(
                    rand(time * 1.3 + gl_VertexID), rand(time * 3.4 + gl_VertexID),
                    rand(time * 2.0 + gl_VertexID)
                );
            }
            ''',
            varyings=['out_pos', 'out_vel', 'out_color'],
        )
        self.gpu_emitter_vao = self.ctx._vertex_array(self.gpu_emitter_prog, [])

        # Query object to inspect render calls
        self.query = self.ctx.query(primitives=True)

    def render(self, time, frame_time):
        self.transform['ft'].value = max(frame_time, 0.02)

        if not self.mouse_control:
            self.move_mouse(time)

        # Cycle emit methods per frame
        method = self.wnd.frames % 3 + 1
        # ---> HARDCODE METHOD HERE <---
        # method = 3
        if method == 1:
            self.emit_cpu_simple(time, frame_time)
        elif method == 2:
            self.emit_cpu_buffer(time, frame_time)
        elif method == 3:
            self.emit_gpu(time, frame_time)

        # Swap around objects for next frame
        self.transform_vao1, self.transform_vao2 = self.transform_vao2, self.transform_vao1
        self.render_vao1, self.render_vao2 = self.render_vao2, self.render_vao1
        self.vbo1, self.vbo2 = self.vbo2, self.vbo1

        err = self.ctx.error 
        if err != "GL_NO_ERROR":
            print(err)

    def emit_cpu_simple(self, time, frame_time):
        """Method #1
        Emit new particles simply by writing new particles to
        to the end of the buffer. This will most likely cause
        the entire buffer to be re-written to the gpu, but is
        probably acceptable for smaller particle amounts.
        """
        # Transform all particles recoding how many elements were emitted by geometry shader
        with self.query:
            self.transform_vao1.transform(self.vbo2, moderngl.POINTS, vertices=self.active_particles)

        # Emit new particles if needed simply by writing to the end of the buffer (cpu)
        emit_count = min(self.N - self.query.primitives, self.max_emit_count)
        if emit_count > 0:
            self.vbo2.write(
                np.fromiter(self.gen_particles(emit_count), 'f4', count=emit_count * self.floats),
                offset=self.query.primitives * self.stride,
            )

        self.active_particles = self.query.primitives + emit_count
        self.render_vao2.render(moderngl.POINTS, vertices=self.active_particles)

    def emit_cpu_buffer(self, time, frame_time):
        """Method #2
        Emit new particles by writing new ones with the cpu
        to a separate emit buffer. This buffer can be smaller
        than the main particle buffer meaning less buffer
        data to stream to the gpu per frame.
        """
        # Transform all particles recoding how many elements were emitted by geometry shader
        with self.query:
            self.transform_vao1.transform(self.vbo2, moderngl.POINTS, vertices=self.active_particles)

        emit_count = min(self.N - self.query.primitives, self.emit_buffer_elements, self.max_emit_count)
        if emit_count > 0:
            self.emit_buffer.write(
                np.fromiter(self.gen_particles(emit_count), 'f4', count=emit_count * self.floats)
            )
            self.emit_buffer_vao.transform(
                self.vbo2,
                vertices=emit_count,
                buffer_offset=self.query.primitives * self.stride,
            )

        self.active_particles = self.query.primitives + emit_count
        self.render_vao2.render(moderngl.POINTS, vertices=self.active_particles)

    def emit_gpu(self, time, frame_time):
        """Method #3
        Emit new particles using a shader.
        """
        # Transform all particles recoding how many elements were emitted by geometry shader
        with self.query:
            self.transform_vao1.transform(self.vbo2, moderngl.POINTS, vertices=self.active_particles)

        emit_count = min(self.N - self.query.primitives, self.emit_buffer_elements, self.max_emit_count)
        if emit_count > 0:
            self.gpu_emitter_prog['mouse_pos'].value = self.mouse_pos
            self.gpu_emitter_prog['mouse_vel'].value = self.mouse_velocity
            self.gpu_emitter_prog['time'].value = max(time, 0)
            self.gpu_emitter_vao.transform(
                self.vbo2,
                vertices=emit_count,
                buffer_offset=self.query.primitives * self.stride,
            )

        self.active_particles = self.query.primitives + emit_count
        self.render_vao2.render(moderngl.POINTS, vertices=self.active_particles)

    def gen_particles(self, n):
        for _ in range(n):
            # Current mouse position (2 floats)
            yield self.mouse_pos[0]
            yield self.mouse_pos[1]
            # Random velocity (2 floats)
            a = np.random.uniform(0.0, np.pi * 2.0)
            r = np.random.uniform(0.1, 0.9)
            yield np.cos(a) * r + self.mouse_velocity[0]
            yield np.sin(a) * r + self.mouse_velocity[1]
            # Random color (3 floats)
            yield np.random.uniform(0.0, 1.0)
            yield np.random.uniform(0.0, 1.0)
            yield np.random.uniform(0.0, 1.0)

    @property
    def projection(self):
        return matrix44.create_orthogonal_projection(
            -self.wnd.aspect_ratio, self.wnd.aspect_ratio,
            -1, 1,
            -1, 1,
            dtype='f4',
        )

    def resize(self, width, height):
        """Handle window resizing"""
        self.prog['projection'].write(self.projection)

    def mouse_position_event(self, x, y, dx, dy):
        if not self.mouse_control:
            return

        self.mouse_pos = (
            (x / self.wnd.width * 2 - 1) * self.wnd.aspect_ratio,
            -(y / self.wnd.height * 2 - 1),
        )
        self.mouse_velocity = dx / 25, -dy / 25

    def move_mouse(self, time):
        new_pos = (
            np.sin(time * 2.0) * self.wnd.aspect_ratio * 0.9,
            np.cos(time * 2.0) * 0.15,
        )
        self.mouse_velocity = (
            (new_pos[0] - self.mouse_pos[0]) * 10,
            (new_pos[1] - self.mouse_pos[1]) * 10,
        )
        self.mouse_pos = new_pos

    def key_event(self, key, action, modifiers):
        keys = self.wnd.keys

        if action == keys.ACTION_PRESS and key == keys.F1:
            screenshot.create(self.wnd.fbo)


if __name__ == '__main__':
    Particles.run()



import numpy as np
from PIL import Image

import _simple_2d_example


class Example(_simple_2d_example.Example):
    title = 'Texture Triangle'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # the shader program is inherited
        # self.prog = self.ctx.program(...)
        self.prog['RenderMode'].value = self.TEXTURE_MODE

        vertices = np.array([
            # x, y, tx, ty
            0.0, 0.8, 0.5, 1.0,
            -0.6, -0.8, 0.0, 0.0,
            0.6, -0.8, 1.0, 0.0,
        ])

        img = Image.open('examples/data/wood.jpg')
        self.texture = self.ctx.texture(img.size, 3, img.tobytes())
        self.sampler = self.ctx.sampler(texture=self.texture)

        self.vbo = self.ctx.buffer(vertices.astype('f4').tobytes())
        self.vao = self.ctx.simple_vertex_array(self.prog, self.vbo, 'in_vert', 'in_text')

        self.vao.scope = self.ctx.scope(samplers=[
            self.sampler.assign(0),
        ])

    def render(self, time: float, frame_time: float):
        self.ctx.clear(1.0, 1.0, 1.0)
        self.vao.render()


if __name__ == '__main__':
    Example.run()



"""
Demonstrates redering a terrain/height map on the fly without any
pre-generated geometry.
"""

import numpy as np
from pyrr import Matrix44, Matrix33

import moderngl
from ported._example import Example


class HeightmapOnTheFly(Example):
    title = "Heightmap - On the fly"
    gl_version = (3, 3)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader="""
                #version 330

                uniform int dim;
                out vec2 uv;

                void main() {
                    // grid position from gl_VertexID normalized
                    vec2 pos = vec2(gl_VertexID % dim, gl_VertexID / dim) / dim;
                    gl_Position = vec4(pos, 0.0, 1.0);
                }
            """,
            geometry_shader="""
            #version 330

            uniform sampler2D heightmap;

            uniform mat4 projection;
            uniform mat4 modelview;
            uniform mat3 normal_matrix;
            uniform int dim;
            uniform float terrain_size;

            out vec2 g_uv;
            // out vec3 g_pos;
            out vec3 normal;

            layout(points) in;
            layout(triangle_strip, max_vertices = 4) out;

            const float scale = 0.5;
            const float height = -0.15;

            float calculateHeight(float h) {
                return h * scale + height;
            }

            vec3 calculateNormal(vec2 uv, float step, float size) {
                float hl = calculateHeight(texture(heightmap, uv + vec2(-step, 0.0)).r);
                float hr = calculateHeight(texture(heightmap, uv + vec2(step, 0.0)).r);
                float hu = calculateHeight(texture(heightmap, uv + vec2(0.0, step)).r);
                float hd = calculateHeight(texture(heightmap, uv + vec2(0.0, -step)).r);
                return normalize(vec3(hl - hr, hd - hu, size));
            }

            void main() {
                // width and height of a quad
                float size = terrain_size / dim;

                // lower left corner of the quad
                vec2 pos = gl_in[0].gl_Position.xy * terrain_size - terrain_size / 2.0;
                vec2 uv = gl_in[0].gl_Position.xy;
                float uv_step = 1.0 / dim;

                // Calculate mvp
                mat4 mvp = projection * modelview;

                // Read heights for each corner
                vec2 uv1 = uv + vec2(0.0, uv_step);
                float h1 = calculateHeight(texture(heightmap, uv1).r);

                vec2 uv2 = uv;
                float h2 = calculateHeight(texture(heightmap, uv2).r);

                vec2 uv3 = uv + vec2(uv_step, uv_step);
                float h3 = calculateHeight(texture(heightmap, uv3).r);

                vec2 uv4 = uv + vec2(uv_step, 0.0);
                float h4 = calculateHeight(texture(heightmap, uv4).r);

                // Upper left
                vec4 pos1 = vec4(pos + vec2(0.0, size), h1, 1.0);
                gl_Position = mvp * pos1;
                g_uv = uv1;
                normal = normal_matrix * calculateNormal(uv1, uv_step, size);
                // g_pos = (modelview * pos1).xyz;
                EmitVertex();

                // Lower left
                vec4 pos2 = vec4(pos, h2, 1.0);
                gl_Position = mvp * pos2;
                g_uv = uv2;
                normal = normal_matrix * calculateNormal(uv2, uv_step, size);
                // g_pos = (modelview * pos2).xyz;
                EmitVertex();

                // Upper right
                vec4 pos3 = vec4(pos + vec2(size, size), h3, 1.0);
                gl_Position = mvp * pos3;
                g_uv = uv3;
                normal = normal_matrix * calculateNormal(uv3, uv_step, size);
                // g_pos = (modelview * pos3).xyz;
                EmitVertex();

                // Lower right
                vec4 pos4 = vec4(pos + vec2(size, 0.0), h4, 1.0);
                gl_Position = mvp * pos4;
                g_uv = uv4;
                normal = normal_matrix * calculateNormal(uv4, uv_step, size);
                // g_pos = (modelview * pos4).xyz;
                EmitVertex();

                EndPrimitive();
            }
            """,
            fragment_shader="""
                #version 330

                uniform sampler2D heightmap;
                out vec4 fragColor;
                in vec2 g_uv;
                // in vec3 g_pos;
                in vec3 normal;

                void main() {
                    // vec3 normal = normalize(cross(dFdx(g_pos), dFdy(g_pos)));
                    float l = abs(dot(vec3(0, 0, 1), normal));

                    // fragColor = vec4(vec3(texture(heightmap, g_uv).r) * l, 1.0);
                    // fragColor = vec4(normal * l, 1.0);
                    fragColor = vec4(vec3(1.0) * l, 1.0);
                }
            """,
        )
        self.heightmap = self.load_texture_2d('heightmap_detailed.png')
        self.heightmap.repeat_x = False
        self.heightmap.repeat_y = False
        self.dim = self.heightmap.width
        self.vao = self.ctx.vertex_array(self.prog, [])

        projection = Matrix44.perspective_projection(45.0, self.aspect_ratio, 0.1, 1000.0, dtype='f4')
        self.prog['projection'].write(projection)
        self.prog['dim'] = self.dim - 1
        self.prog['terrain_size'] = 1.0

    def render(self, time, frame_time):
        self.ctx.clear()
        self.ctx.enable(moderngl.DEPTH_TEST | moderngl.CULL_FACE)
        angle = time * 0.2

        lookat = Matrix44.look_at(
            (np.cos(angle), np.sin(angle), 0.4),
            (0.0, 0.0, 0.0),
            (0.0, 0.0, 1.0),
            dtype='f4',
        )
        normal_matrix = Matrix33.from_matrix44(lookat).inverse.transpose()

        self.prog['modelview'].write(lookat)
        self.prog['normal_matrix'].write(normal_matrix.astype('f4').tobytes())
        self.heightmap.use(0)
        self.vao.render(moderngl.POINTS, vertices=(self.dim - 1) ** 2)


if __name__ == '__main__':
    HeightmapOnTheFly.run()



import struct

from PyQt5 import QtOpenGL, QtWidgets

import ModernGL
from pyrr import Matrix44


class QGLControllerWidget(QtOpenGL.QGLWidget):
    def __init__(self):
        fmt = QtOpenGL.QGLFormat()
        fmt.setVersion(3, 3)
        fmt.setProfile(QtOpenGL.QGLFormat.CoreProfile)
        fmt.setSampleBuffers(True)
        super(QGLControllerWidget, self).__init__(fmt, None)

    def initializeGL(self):
        self.ctx = ModernGL.create_context()

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                uniform mat4 Mvp;

                in vec3 in_vert;
                in vec3 in_color;

                out vec3 v_color;

                void main() {
                    v_color = in_color;
                    gl_Position = Mvp * vec4(in_vert, 1.0);
                }
            '''),
            fragment_shader='''
                #version 330

                in vec3 v_color;
                out vec4 f_color;

                void main() {
                    f_color = vec4(v_color, 1.0);
                }
            '''),
        ])

        self.mvp = self.prog.uniforms['Mvp']

        grid = bytearray()

        for i in range(0, 32 + 1):
            grid += struct.pack('6f', i - 16.0, -16.0, 0.0, 0.0, 0.0, 0.0)
            grid += struct.pack('6f', i - 16.0, 16.0, 0.0, 0.0, 0.0, 0.0)
            grid += struct.pack('6f', -16.0, i - 16.0, 0.0, 0.0, 0.0, 0.0)
            grid += struct.pack('6f', 16.0, i - 16.0, 0.0, 0.0, 0.0, 0.0)

        self.vbo = self.ctx.buffer(grid)
        self.vao = self.ctx.simple_vertex_array(self.prog, self.vbo, ['in_vert', 'in_color'])

    def paintGL(self):
        width, height = self.width(), self.height()
        self.ctx.viewport = (0, 0, width, height)
        self.ctx.clear(0.9, 0.9, 0.9)
        self.ctx.enable(ModernGL.DEPTH_TEST)

        proj = Matrix44.perspective_projection(45.0, width / height, 0.1, 1000.0)
        lookat = Matrix44.look_at(
            (40.0, 30.0, 20.0),
            (0.0, 0.0, 0.0),
            (0.0, 0.0, 1.0),
        )

        self.mvp.write((proj * lookat).astype('float32').tobytes())
        self.vao.render(ModernGL.LINES)


app = QtWidgets.QApplication([])
window = QGLControllerWidget()
window.move(QtWidgets.QDesktopWidget().rect().center() - window.rect().center())
window.show()
app.exec_()



import struct

import ModernGL
from PyQt5 import QtCore, QtOpenGL, QtWidgets


class QGLControllerWidget(QtOpenGL.QGLWidget):
    def __init__(self):
        fmt = QtOpenGL.QGLFormat()
        fmt.setVersion(3, 3)
        fmt.setProfile(QtOpenGL.QGLFormat.CoreProfile)
        fmt.setSampleBuffers(True)
        self.timer = QtCore.QElapsedTimer()
        self.timer.restart()
        super(QGLControllerWidget, self).__init__(fmt, None)

    def initializeGL(self):
        self.ctx = ModernGL.create_context()

        prog = self.ctx.program(
            vertex_shader='''
                #version 330

                in vec2 vert;

                in vec3 vert_color;
                out vec3 frag_color;

                uniform vec2 scale;
                uniform float rotation;

                void main() {
                    frag_color = vert_color;
                    mat2 rot = mat2(
                        cos(rotation), sin(rotation),
                        -sin(rotation), cos(rotation)
                    );
                    gl_Position = vec4((rot * vert) * scale, 0.0, 1.0);
                }
            '''),
            fragment_shader='''
                #version 330

                in vec3 frag_color;
                out vec4 color;

                void main() {
                    color = vec4(frag_color, 1.0);
                }
            '''),
        ])

        self.scale = prog.uniforms['scale']
        self.rotation = prog.uniforms['rotation']

        vbo = self.ctx.buffer(struct.pack(
            '15f',
            1.0, 0.0, 1.0, 0.0, 0.0,
            -0.5, 0.86, 0.0, 1.0, 0.0,
            -0.5, -0.86, 0.0, 0.0, 1.0,
        ))

        self.vao = self.ctx.simple_vertex_array(prog, vbo, ['vert', 'vert_color'])

    def paintGL(self):
        self.ctx.viewport = (0, 0, self.width(), self.height())
        self.ctx.clear(0.9, 0.9, 0.9)
        self.scale.value = (self.height() / self.width() * 0.75, 0.75)
        self.rotation.value = self.timer.elapsed() / 1000
        self.vao.render()
        self.ctx.finish()
        self.update()


app = QtWidgets.QApplication([])
window = QGLControllerWidget()
window.move(QtWidgets.QDesktopWidget().rect().center() - window.rect().center())
window.show()
app.exec_()



"""

"""
import numpy as np

import moderngl as mgl


def test_cull_face(standalone_context,
                   prog_render_depth_pass,
                   vbo_triangle,
                   fbo_with_rasterised_triangle,
                   np_triangle_rasterised):
    ctx = standalone_context

    size = (16,) * 2

    ctx.enable(mgl.DEPTH_TEST)
    ctx.enable(mgl.CULL_FACE)

    def _do_test_cull_face(cull_face: str, expected_results: np.array):
        ctx.cull_face = cull_face
        fbo_depth, tex_depth = fbo_with_rasterised_triangle(prog_render_depth_pass, size)
        depth_from_dbo = np.frombuffer(tex_depth.read(), dtype=np.dtype('f4')).reshape(size[::-1])
        np.testing.assert_array_almost_equal(depth_from_dbo, expected_results)

    ############################################################################
    # EXPECTED DATAS                                                                                                   #
    # It should have 0.5's where the triangle lies.                                                                    #
    ############################################################################
    np_triangle_raster = np_triangle_rasterised(size)
    ############################################################################
    # cull face is set to back, only one 'front' triangle in the scene => no culling
    _do_test_cull_face('back', np_triangle_raster)

    ############################################################################
    # EXPECTED DATAS                                                                                                   #
    # It should have 1.0's everywhere.                                                                    #
    ############################################################################
    np_clear_depth = np.full(size, 1.0)
    ############################################################################
    # cull face is set to front, only one 'front' triangle in the scene => the triangle is culled
    _do_test_cull_face('front', np_clear_depth)

    # cull face is set to front_and_back, only one 'front' triangle in the scene => the triangle is culled
    _do_test_cull_face('front_and_back', np_clear_depth)

    ctx.enable_only(mgl.NOTHING)



import numpy as np

import _simple_2d_example


class Example(_simple_2d_example.Example):
    title = 'Color Triangle'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # the shader program is inherited
        # self.prog = self.ctx.program(...)
        self.prog['RenderMode'] = self.ATTRIB_COLOR_MODE

        vertices = np.array([
            # x, y, red, green, blue
            0.0, 0.8, 1.0, 0.0, 0.0,
            -0.6, -0.8, 0.0, 1.0, 0.0,
            0.6, -0.8, 0.0, 0.0, 1.0,
        ])

        self.vbo = self.ctx.buffer(vertices.astype('f4').tobytes())
        self.vao = self.ctx.vertex_array(self.prog, self.vbo, 'in_vert', 'in_color')

    def render(self, time: float, frame_time: float):
        self.ctx.screen.clear(color=(1.0, 1.0, 1.0))
        self.vao.render()


if __name__ == '__main__':
    Example.run()



from PIL import Image
import numpy as np
from objloader import Obj

import _simple_3d_example
import moderngl


class Example(_simple_3d_example.Example):
    title = 'Hello World'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # the shader program is inherited
        # self.prog = self.ctx.program(...)
        self.prog['RenderMode'] = self.TEXTURE_WITH_LIGHT_MODE

        obj = Obj.open('examples/data/crate.obj')
        img = Image.open('examples/data/crate.png')
        self.texture = self.ctx.texture(img.size, 3, img.tobytes('raw', 'RGB', 0, -1))
        self.sampler = self.ctx.sampler(texture=self.texture)

        self.vbo = self.ctx.buffer(obj.pack('vx vy vz nx ny nz tx ty'))
        self.vao = self.ctx.vertex_array(self.prog, self.vbo, 'in_vert', 'in_norm', 'in_text')
        self.vao.scope = self.ctx.scope(enable=self.ctx.DEPTH_TEST, samplers=[
            self.sampler.assign(0),
        ])

    def render(self, time: float, frame_time: float):
        self.ctx.screen.clear(color=(1.0, 1.0, 1.0))
        self.set_camera(fov=60.0, eye=(4, 3, 2), target=(0, 0, 0))
        self.vao.render()


if __name__ == '__main__':
    Example.run()



import numpy as np

import moderngl
from ported._example import Example


def particle():
    a = np.random.uniform(0.0, np.pi * 2.0)
    r = np.random.uniform(0.0, 0.001)
    return np.array([0.0, 0.0, np.cos(a) * r - 0.003, np.sin(a) * r - 0.008]).astype('f4')


class Particles(Example):
    title = "Particle System"
    gl_version = (3, 3)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                in vec2 in_vert;

                void main() {
                    gl_Position = vec4(in_vert, 0.0, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330

                out vec4 f_color;

                void main() {
                    f_color = vec4(0.30, 0.50, 1.00, 1.0);
                }
            ''',
        )

        self.transform = self.ctx.program(
            vertex_shader='''
            #version 330

            uniform vec2 Acc;

            in vec2 in_pos;
            in vec2 in_prev;

            out vec2 out_pos;
            out vec2 out_prev;

            void main() {
                vec2 velocity = in_pos - in_prev;
                out_pos = in_pos + velocity + Acc;
                out_prev = in_pos;
            }
        ''',
            varyings=['out_pos', 'out_prev']
        )

        self.acc = self.transform['Acc']
        self.acc.value = (0.0, -0.0001)

        self.vbo1 = self.ctx.buffer(b''.join(particle() for i in range(1024)))
        self.vbo2 = self.ctx.buffer(reserve=self.vbo1.size)

        self.vao1 = self.ctx.simple_vertex_array(self.transform, self.vbo1, 'in_pos', 'in_prev')
        self.vao2 = self.ctx.simple_vertex_array(self.transform, self.vbo2, 'in_pos', 'in_prev')

        self.render_vao = self.ctx.vertex_array(self.prog, [
            (self.vbo1, '2f 2x4', 'in_vert'),
        ])

        self.idx = 0

    def render(self, time, frame_time):
        self.ctx.clear(1.0, 1.0, 1.0)
        self.ctx.point_size = 2.0

        for i in range(8):
            self.vbo1.write(particle(), offset=self.idx * 16)
            self.idx = (self.idx + 1) % 1024

        self.render_vao.render(moderngl.POINTS, 1024)
        self.vao1.transform(self.vbo2, moderngl.POINTS, 1024)
        self.ctx.copy_buffer(self.vbo1, self.vbo2)


if __name__ == '__main__':
    Particles.run()



import unittest

from common import get_context


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = get_context()

    def test_1(self):
        fbo1 = self.ctx.simple_framebuffer((4, 4))
        fbo2 = self.ctx.simple_framebuffer((4, 4))

        fbo1.clear(0.0, 0.0, 0.0, 0.0)

        self.assertEqual(fbo1.read(), b'\x00\x00\x00' * 4 * 4)

        fbo1.clear(1.0, 0.0, 0.0, 0.0, viewport=(2, 2))
        fbo1.clear(0.0, 1.0, 0.0, 0.0, viewport=(2, 0, 2, 2))
        fbo1.clear(0.0, 0.0, 1.0, 0.0, viewport=(0, 2, 2, 2))

        self.assertEqual(fbo1.read((2, 2)), b'\xff\x00\x00' * 2 * 2)
        self.assertEqual(fbo1.read((2, 0, 2, 2)), b'\x00\xff\x00' * 2 * 2)
        self.assertEqual(fbo1.read((0, 2, 2, 2)), b'\x00\x00\xff' * 2 * 2)
        self.assertEqual(fbo1.read((2, 2, 2, 2)), b'\x00\x00\x00' * 2 * 2)

        self.ctx.copy_framebuffer(fbo2, fbo1)

        self.assertEqual(fbo2.read((2, 2)), b'\xff\x00\x00' * 2 * 2)
        self.assertEqual(fbo2.read((2, 0, 2, 2)), b'\x00\xff\x00' * 2 * 2)
        self.assertEqual(fbo2.read((0, 2, 2, 2)), b'\x00\x00\xff' * 2 * 2)
        self.assertEqual(fbo2.read((2, 2, 2, 2)), b'\x00\x00\x00' * 2 * 2)


if __name__ == '__main__':
    unittest.main()



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import struct

# import GLWindow
# import ModernGL
# from pyrr import Matrix44

# wnd = GLWindow.create_window()
# ctx = ModernGL.create_context()

# prog = ctx.program(
#     ctx.vertex_shader('''
#         #version 330

#         uniform mat4 Mvp;

#         in vec3 in_vert;
#         in vec3 in_color;

#         out vec3 v_color;

#         void main() {
#             v_color = in_color;
#             gl_Position = Mvp * vec4(in_vert, 1.0);
#         }
#     '''),
#     ctx.fragment_shader('''
#         #version 330

#         in vec3 v_color;
#         out vec4 f_color;

#         void main() {
#             f_color = vec4(v_color, 1.0);
#         }
#     '''),
# ])

# mvp = prog.uniforms['Mvp']

# grid = bytearray()

# for i in range(0, 32 + 1):
#     grid += struct.pack('6f', i - 16.0, -16.0, 0.0, 0.0, 0.0, 0.0)
#     grid += struct.pack('6f', i - 16.0, 16.0, 0.0, 0.0, 0.0, 0.0)
#     grid += struct.pack('6f', -16.0, i - 16.0, 0.0, 0.0, 0.0, 0.0)
#     grid += struct.pack('6f', 16.0, i - 16.0, 0.0, 0.0, 0.0, 0.0)

# vbo = ctx.buffer(grid)
# vao = ctx.simple_vertex_array(prog, vbo, ['in_vert', 'in_color'])

# while wnd.update():
#     ctx.viewport = wnd.viewport
#     ctx.clear(0.9, 0.9, 0.9)
#     ctx.enable(ModernGL.DEPTH_TEST)

#     proj = Matrix44.perspective_projection(45.0, wnd.ratio, 0.1, 1000.0)
#     lookat = Matrix44.look_at(
#         (40.0, 30.0, 20.0),
#         (0.0, 0.0, 0.0),
#         (0.0, 0.0, 1.0),
#     )

#     mvp.write((proj * lookat).astype('float32').tobytes())
#     vao.render(ModernGL.LINES)



import unittest

import moderngl
from common import get_context


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = get_context()
        cls.vendor = cls.ctx.info['GL_VENDOR']

    def test_texture_3d_create_1(self):
        self.ctx.texture3d((8, 8, 8), 3)

    def test_texture_3d_create_2(self):
        pixels = b'\x10\x20\x30' * 8 * 8 * 8
        self.ctx.texture3d((8, 8, 8), 3, pixels)

    def test_texture_3d_create_string(self):
        pixels = 'abc' * 8 * 8
        with self.assertRaises(Exception):
            self.ctx.texture3d((8, 8, 8), 3, pixels)

    def test_texture_3d_create_wrong_size(self):
        with self.assertRaises(Exception):
            self.ctx.texture3d((8, 8), 3)

    def test_texture_3d_get_swizzle(self):
        tex = self.ctx.texture3d((8, 8, 8), 4)
        self.assertEqual(tex.swizzle, 'RGBA')

    def test_texture_3d_swizzle_1(self):
        tex = self.ctx.texture3d((8, 8, 8), 4)
        tex.swizzle = 'argb'
        self.assertEqual(tex.swizzle, 'ARGB')

    def test_texture_3d_swizzle_2(self):
        tex = self.ctx.texture3d((8, 8, 8), 1)
        tex.swizzle = 'RRRR'
        self.assertEqual(tex.swizzle, 'RRRR')

    def test_texture_3d_swizzle_3(self):
        tex = self.ctx.texture3d((8, 8, 8), 2)
        tex.swizzle = '01RG'
        self.assertEqual(tex.swizzle, '01RG')

    def test_texture_3d_read(self):
        pixels = b'\x10\x20\x30' * 8 * 8 * 8
        tex = self.ctx.texture3d((8, 8, 8), 3, pixels)
        self.assertEqual(tex.read(), pixels)

    def test_texture_3d_read_into(self):
        pixels = b'\x10\x20\x30' * 8 * 8 * 8
        tex = self.ctx.texture3d((8, 8, 8), 3, pixels)
        buf = bytearray(8 * 8 * 8 * 3)
        tex.read_into(buf)
        self.assertEqual(bytes(buf), pixels)

    def test_texture_3d_read_into_pbo(self):
        if self.vendor.startswith('Intel'):
            raise unittest.SkipTest('')

        pixels = b'\x10\x20\x30' * 8 * 8 * 8
        tex = self.ctx.texture3d((8, 8, 8), 3, pixels)
        buf = self.ctx.buffer(reserve=8 * 8 * 8 * 3)
        tex.read_into(buf)
        self.assertEqual(buf.read(), pixels)

    def test_texture_3d_write_1(self):
        pixels1 = b'\x00\x00\x00' * 8 * 8 * 8
        pixels2 = b'\xff\xff\xff' * 8 * 8 * 8

        tex = self.ctx.texture3d((8, 8, 8), 3, pixels1)
        self.assertEqual(tex.read(), pixels1)

        tex.write(pixels2)
        self.assertEqual(tex.read(), pixels2)

    def test_texture_3d_write_2(self):
        pixels1 = b'\x00\x00\x00' * 8 * 8 * 8
        pixels2 = b'\xff\xff\xff' * 6 * 6 * 6
        pixels3 = b'\x10\x20\x30'

        tex = self.ctx.texture3d((8, 8, 8), 3)

        tex.write(pixels1, viewport=(0, 0, 0, 8, 8, 8))
        tex.write(pixels2, viewport=(1, 1, 1, 6, 6, 6))
        tex.write(pixels3, viewport=(2, 4, 6, 1, 1, 1))

        def pixel(x, y, z):
            if x == 2 and y == 4 and z == 6:
                return b'\x10\x20\x30'

            if x == 0 or y == 0 or z == 0 or x == 7 or y == 7 or z == 7:
                return b'\x00\x00\x00'

            return b'\xff\xff\xff'

        expectation = b''.join(pixel(x, y, z) for z in range(8) for y in range(8) for x in range(8))
        self.assertEqual(tex.read(), expectation)

    def test_texture_default_filter(self):
        """Ensure default filter is correct"""
        # Float types
        for dtype in ["f1", "f2", "f4"]:
            texture = self.ctx.texture3d((10, 10, 10), 4, dtype=dtype)
            self.assertEqual(texture.filter, (moderngl.LINEAR, moderngl.LINEAR))

        for dtype in ["u1", "u2", "u4", "i1", "i2", "i4"]:
            texture = self.ctx.texture3d((10, 10, 10), 4, dtype=dtype)
            self.assertEqual(texture.filter, (moderngl.NEAREST, moderngl.NEAREST))

    def test_bind_to_image(self):
        tex = self.ctx.texture3d((4, 4, 4), 4)
        tex.bind_to_image(0, read=True, write=False)
        tex.bind_to_image(1, read=False, write=True)


if __name__ == '__main__':
    unittest.main()



import unittest

from common import get_context


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = get_context()

    def test_1(self):
        pass


if __name__ == '__main__':
    unittest.main()



import numpy as np
import moderngl

import _simple_3d_example


def grid(size, steps):
    u = np.repeat(np.linspace(-size, size, steps), 2)
    v = np.tile([-size, size], steps)
    w = np.zeros(steps * 2)
    return np.concatenate([np.dstack([u, v, w]), np.dstack([v, u, w])])


class Example(_simple_3d_example.Example):
    title = 'Hello World'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # the shader program is inherited
        # self.prog = self.ctx.program(...)

        self.vbo = self.ctx.buffer(grid(2.0, 10).astype('f4').tobytes())
        self.vao = self.ctx.vertex_array(self.prog, self.vbo, 'in_vert')

    def render(self, time: float, frame_time: float):
        self.ctx.screen.clear(color=(1.0, 1.0, 1.0))
        self.prog['Color'] = (0.0, 0.0, 0.0)
        self.set_camera(fov=60.0, eye=(4, 3, 2), target=(0, 0, 0))
        self.vao.render(moderngl.LINES)


if __name__ == '__main__':
    Example.run()



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import struct
# import tkinter

# import ModernGL
# import ModernGL.tk

# ctx = ModernGL.create_standalone_context()

# prog = ctx.program(
#     ctx.vertex_shader('''
#         #version 330

#         in vec2 vert;

#         in vec3 vert_color;
#         out vec3 frag_color;

#         uniform vec2 scale;
#         uniform float rotation;

#         void main() {
#             frag_color = vert_color;
#             mat2 rot = mat2(
#                 cos(rotation), sin(rotation),
#                 -sin(rotation), cos(rotation)
#             );
#             gl_Position = vec4((rot * vert) * scale, 0.0, 1.0);
#         }
#     '''),
#     ctx.fragment_shader('''
#         #version 330

#         in vec3 frag_color;
#         out vec4 color;

#         void main() {
#             color = vec4(frag_color, 1.0);
#         }
#     '''),
# ])

# # Uniforms

# scale = prog.uniforms['scale']
# rotation = prog.uniforms['rotation']

# width, height = 512, 512
# scale.value = (height / width * 0.75, 0.75)

# # Buffer

# vbo = ctx.buffer(struct.pack(
#     '15f',

#     1.0, 0.0,
#     1.0, 0.0, 0.0,

#     -0.5, 0.86,
#     0.0, 1.0, 0.0,

#     -0.5, -0.86,
#     0.0, 0.0, 1.0,
# ))

# # Put everything together

# vao = ctx.simple_vertex_array(prog, vbo, ['vert', 'vert_color'])

# # Main loop


# size = (512, 512)
# color_rbo = ctx.renderbuffer(size)
# depth_rbo = ctx.depth_renderbuffer(size)
# fbo = ctx.framebuffer(color_rbo, depth_rbo)

# root = tkinter.Tk()

# # frame = tkinter.Frame(root, width=512, height=512, bg="", colormap="new")
# frame = ModernGL.tk.ModernGLCanvas(root, 512, 512)
# frame.pack()

# # connector = ModernGL.tk.connect(frame)
# R = tkinter.Scale(root, from_=0, to=255, orient=tkinter.HORIZONTAL, length=300)
# R.pack()

# k = 1


# def task():
#     global k
#     k += 0.02
#     fbo.use()
#     ctx.viewport = (0, 0, 512, 512)
#     ctx.clear(0.9, 0.9, 0.9)
#     rotation.value = k
#     vao.render()

#     fbo.read_into(frame.pixels, components=3)
#     frame.update()

#     root.after(1, task)


# root.after(1, task)
# tkinter.mainloop()



import unittest

from common import get_context


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = get_context()

    # def test_viewport(self):
    #     self.ctx.screen.use()
    #     self.ctx.viewport = (1, 2, 3, 4)
    #     self.assertTupleEqual(self.ctx.viewport, (1, 2, 3, 4))
    #     self.assertTupleEqual(self.ctx.screen.viewport, (1, 2, 3, 4))
    #     fbo = self.ctx.simple_framebuffer((10, 10))
    #     fbo.use()
    #     self.assertTupleEqual(self.ctx.viewport, (0, 0, 10, 10))
    #     self.assertTupleEqual(fbo.viewport, (0, 0, 10, 10))
    #     self.ctx.viewport = (2, 3, 7, 8)
    #     self.assertTupleEqual(fbo.viewport, (2, 3, 7, 8))
    #     self.assertTupleEqual(self.ctx.viewport, (2, 3, 7, 8))
    #     self.assertTupleEqual(self.ctx.screen.viewport, (1, 2, 3, 4))
    #     self.ctx.screen.viewport = (3, 3, 3, 3)
    #     self.assertTupleEqual(fbo.viewport, (2, 3, 7, 8))
    #     self.assertTupleEqual(self.ctx.viewport, (2, 3, 7, 8))
    #     self.assertTupleEqual(self.ctx.screen.viewport, (3, 3, 3, 3))
    #     self.ctx.screen.use()
    #     self.assertTupleEqual(self.ctx.viewport, (3, 3, 3, 3))
    #     fbo.viewport = (0, 0, 7, 7)
    #     self.assertTupleEqual(self.ctx.viewport, (3, 3, 3, 3))
    #     self.assertTupleEqual(self.ctx.screen.viewport, (3, 3, 3, 3))
    #     self.ctx.viewport = (3, 5, 7, 9)
    #     self.assertTupleEqual(fbo.viewport, (0, 0, 7, 7))
    #     self.assertTupleEqual(self.ctx.viewport, (3, 5, 7, 9))
    #     self.assertTupleEqual(self.ctx.screen.viewport, (3, 5, 7, 9))
    #     fbo.use()
    #     self.assertTupleEqual(self.ctx.viewport, (0, 0, 7, 7))
    #     self.ctx.screen.use()


if __name__ == '__main__':
    unittest.main()



import GLWindow
import moderngl
from PIL import Image, ImageDraw, ImageFont
import numpy as np

wnd = GLWindow.create_window()
ctx = moderngl.create_context()

prog = ctx.program(
    vertex_shader='''
        #version 330

        in vec2 in_vert;
        in vec3 in_text;

        out vec3 v_text;

        void main() {
            v_text = in_text;
            gl_Position = vec4(in_vert, 0.0, 1.0);
        }
    ''',
    fragment_shader='''
        #version 330

        uniform sampler2DArray Texture;

        in vec3 v_text;

        out vec4 f_color;

        void main() {
            f_color = texture(Texture, v_text);
        }
    ''',
)

vbo = ctx.buffer(np.array([
    (
        -0.3, -0.3 + i * 0.2 - 0.2, 0.0, 0.0, i,
        0.3, -0.3 + i * 0.2 - 0.2, 1.0, 0.0, i,
        0.0, 0.3 + i * 0.2 - 0.2, 0.5, 1.0, i,
    ) for i in range(3)
], dtype='f4'))

images = [
    Image.new('RGB', (256, 256), 'red'),
    Image.new('RGB', (256, 256), 'green'),
    Image.new('RGB', (256, 256), 'blue'),
]

merged = b''.join(img.tobytes() for img in images)

tex = ctx.texture_array((256, 256, 3), 3, merged)
tex.use()

vao = ctx.simple_vertex_array(prog, vbo, 'in_vert', 'in_text')

while wnd.update():
    ctx.clear(1.0, 1.0, 1.0, 1.0)
    vao.render()



import numpy as np
import pymunk
from pymunk import Vec2d

import moderngl
from ported._example import Example


class PymunkExample(Example):
    title = "Using PyMunk"
    gl_version = (3, 3)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                uniform vec4 Camera;

                // Per vertex
                in vec2 in_vert;
                in vec2 in_texture;

                // Per instance
                in vec3 in_pos;
                in vec2 in_size;
                in vec4 in_tint;

                out vec2 v_vert;
                out vec2 v_texture;
                out vec4 v_tint;

                void main() {
                    mat2 rotate = mat2(
                        cos(in_pos.z), sin(in_pos.z),
                        -sin(in_pos.z), cos(in_pos.z)
                    );
                    v_vert = rotate * (in_vert * in_size) + in_pos.xy;
                    gl_Position = vec4((v_vert - Camera.xy) / Camera.zw, 0.0, 1.0);
                    v_texture = in_texture;
                    v_tint = in_tint;
                }
            ''',
            fragment_shader='''
                #version 330

                uniform sampler2D Texture;

                in vec2 v_vert;
                in vec2 v_texture;
                in vec4 v_tint;

                out vec4 f_color;

                void main() {
                    vec4 tex = texture(Texture, v_texture);
                    vec3 color = tex.rgb * (1.0 - v_tint.a) + v_tint.rgb * v_tint.a;
                    f_color = vec4(color, tex.a);
            }
            ''',
        )

        self.tex1 = self.load_texture_2d('crate.png')
        self.tex1.use(0)

        self.tex2 = self.load_texture_2d('ball.png')
        self.tex2.use(1)

        vertices = np.array([
            -1.0, -1.0, 0.0, 0.0,
            -1.0, 1.0, 0.0, 1.0,
            1.0, -1.0, 1.0, 0.0,
            1.0, 1.0, 1.0, 1.0,
        ])

        vbo1 = self.ctx.buffer(vertices.astype('f4'))

        self.vbo2 = self.ctx.buffer(reserve=1024 * 1024)

        vao_content = [
            (vbo1, '2f 2f', 'in_vert', 'in_texture'),
            (self.vbo2, '3f 2f 4f/i', 'in_pos', 'in_size', 'in_tint'),
        ]

        self.vao = self.ctx.vertex_array(self.prog, vao_content)

        self.space = pymunk.Space()
        self.space.gravity = (0.0, -900.0)

        shape = pymunk.Segment(self.space.static_body, (5, 100), (595, 100), 1.0)
        shape.friction = 1.0
        self.space.add(shape)

        self.bodies = []
        self.balls = []

        for x in range(5):
            for y in range(10):
                size = 20
                mass = 10.0
                moment = pymunk.moment_for_box(mass, (size, size))
                body = pymunk.Body(mass, moment)
                body.position = Vec2d(300 + x * 50, 105 + y * (size + 0.1))
                shape = pymunk.Poly.create_box(body, (size, size))
                shape.friction = 0.3
                self.space.add(body, shape)
                self.bodies.append(body)

    def shoot(self):
        mass = 100
        r = 15
        moment = pymunk.moment_for_circle(mass, 0, r, (0, 0))
        body = pymunk.Body(mass, moment)
        body.position = (0, 165)
        shape = pymunk.Circle(body, r, (0, 0))
        shape.friction = 0.3
        self.space.add(body, shape)
        f = 50000
        body.apply_impulse_at_local_point((f, 0), (0, 0))
        self.balls.append(body)

    def mouse_press_event(self, x: int, y: int, button: int):
        self.shoot()

    def key_event(self, key, action, modifiers):
        if action == self.wnd.keys.ACTION_PRESS:
            self.shoot()

    def render(self, time, frame_time):
        width, height = self.wnd.size

        self.ctx.clear(1.0, 1.0, 1.0)
        self.ctx.enable(moderngl.BLEND)

        for i in range(10):
            self.space.step(1 / 60 / 10)

        self.prog['Camera'].value = (200, 300, width / 2, height / 2)

        bodies = np.array([(b.position.x, b.position.y, b.angle, 10, 10, 1, 1, 1, 0) for b in self.bodies], dtype='f4')
        self.vbo2.write(bodies)
        self.prog['Texture'].value = 0
        self.vao.render(moderngl.TRIANGLE_STRIP, instances=len(self.bodies))

        self.vbo2.orphan()

        balls = np.array([(b.position.x, b.position.y, b.angle, 15, 15, 1, 1, 1, 0) for b in self.balls], dtype='f4')
        self.vbo2.write(balls)
        self.prog['Texture'].value = 1
        self.vao.render(moderngl.TRIANGLE_STRIP, instances=len(self.balls))


if __name__ == '__main__':
    PymunkExample.run()



import struct
import unittest

from common import get_context


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = get_context()
        if 'llvmpipe' in cls.ctx.info['GL_RENDERER']:
            raise unittest.SkipTest('Temp disable for llvmpipe')

        cls.prog = cls.ctx.program(
            vertex_shader='''
                #version 330

                in vec2 in_v;
                out vec2 out_v;

                uniform Block1 {
                    float x;
                };

                uniform Block2 {
                    float y;
                };

                uniform Block3 {
                    float z;
                };

                void main() {
                    out_v = in_v * z + vec2(x, y);
                }
            ''',
            varyings=['out_v']
        )

    def test_1(self):
        buf_v = self.ctx.buffer(struct.pack('2f', 100.0, 1000.0))
        buf_u1 = self.ctx.buffer(struct.pack('f', 9.5))
        buf_u2 = self.ctx.buffer(struct.pack('f', 4.0))
        buf_u3 = self.ctx.buffer(struct.pack('f', 3.0))
        buf_u4 = self.ctx.buffer(struct.pack('f', 0.0))
        buf_r = self.ctx.buffer(reserve=buf_v.size)

        vao = self.ctx.vertex_array(self.prog, [
            (buf_v, '2f', 'in_v'),
        ])

        self.assertIn('Block1', self.prog)
        self.assertIn('Block2', self.prog)
        self.assertIn('Block3', self.prog)

        self.assertNotIn('x', self.prog)
        self.assertNotIn('y', self.prog)
        self.assertNotIn('z', self.prog)

        self.prog['Block1'].binding = 2
        self.prog['Block2'].binding = 4
        self.prog['Block3'].binding = 1

        buf_u1.bind_to_uniform_block(2)
        buf_u2.bind_to_uniform_block(4)
        buf_u4.bind_to_uniform_block(1)

        vao.transform(buf_r)
        a, b = struct.unpack('2f', buf_r.read())
        self.assertAlmostEqual(a, 9.5)
        self.assertAlmostEqual(b, 4.0)

        buf_u3.bind_to_uniform_block(1)

        vao.transform(buf_r)
        a, b = struct.unpack('2f', buf_r.read())
        self.assertAlmostEqual(a, 309.5)
        self.assertAlmostEqual(b, 3004.0)

    def test_2(self):
        min_offset = self.ctx.info['GL_UNIFORM_BUFFER_OFFSET_ALIGNMENT']
        u3_len = (min_offset // 4) + 1
        u3_data = list(range(u3_len))

        buf_v = self.ctx.buffer(struct.pack('2f', 100.0, 1000.0))
        buf_u1 = self.ctx.buffer(struct.pack('f', 9.5))
        buf_u2 = self.ctx.buffer(struct.pack('f', 4.0))
        buf_u3 = self.ctx.buffer(struct.pack('{:d}f'.format(u3_len), *u3_data))
        buf_r = self.ctx.buffer(reserve=buf_v.size)

        vao = self.ctx.vertex_array(self.prog, [
            (buf_v, '2f', 'in_v'),
        ])

        self.prog['Block1'].binding = 2
        self.prog['Block2'].binding = 4
        self.prog['Block3'].binding = 1

        buf_u1.bind_to_uniform_block(2)
        buf_u2.bind_to_uniform_block(4)
        buf_u3.bind_to_uniform_block(1, offset=min_offset)

        vao.transform(buf_r)
        a, b = struct.unpack('2f', buf_r.read())
        self.assertAlmostEqual(a, u3_data[-1] * 100 + 9.5)
        self.assertAlmostEqual(b, u3_data[-1] * 1000 + 4.0)

        buf_u3.bind_to_uniform_block(1, offset=0)

        vao.transform(buf_r)
        a, b = struct.unpack('2f', buf_r.read())
        self.assertAlmostEqual(a, 9.5)
        self.assertAlmostEqual(b, 4.0)


if __name__ == '__main__':
    unittest.main()



import tkinter as tk

import moderngl
import numpy as np

from tkinter_framebuffer import FramebufferImage
from renderer_example import HelloWorld2D, PanTool

ctx = moderngl.create_standalone_context()

canvas = HelloWorld2D(ctx)
pan_tool = PanTool()


def vertices():
    x = np.linspace(-1.0, 1.0, 50)
    y = np.random.rand(50) - 0.5
    r = np.ones(50)
    g = np.zeros(50)
    b = np.zeros(50)
    a = np.ones(50)
    return np.dstack([x, y, r, g, b, a])


verts = vertices()


def update(evt):
    if evt.type == tk.EventType.ButtonPress:
        pan_tool.start_drag(evt.x / size[0], evt.y / size[1])
    if evt.type == tk.EventType.Motion:
        pan_tool.dragging(evt.x / size[0], evt.y / size[1])
    if evt.type == tk.EventType.ButtonRelease:
        pan_tool.stop_drag(evt.x / size[0], evt.y / size[1])
    canvas.pan(pan_tool.value)

    with tkfbo:
        ctx.clear()
        canvas.plot(verts)


size = (512, 512)

root = tk.Tk()
tkfbo = FramebufferImage(root, ctx, size)

lbl = tk.Label(root, image=tkfbo)
lbl.bind("<ButtonPress-1>", update)
lbl.bind("<ButtonRelease-1>", update)
lbl.bind('<Motion>', update)
lbl.pack()

# btn = tk.Button(root, text='Hello', command=update)
# btn.pack()

root.mainloop()



# coding=utf-8
import os
import moderngl as mgl
import numpy as np
import pytest

# https://docs.oracle.com/cd/E19957-01/806-3568/ncg_goldberg.html
# https://en.wikipedia.org/wiki/Machine_epsilon
# binary32	single precision	float	2	24 (one bit is implicit)	2−24 ≈ 5.96e-08	2−23 ≈ 1.19e-07
EPSILON_IEEE_754 = 5.96e-08


@pytest.fixture(scope="session", autouse=True)
def standalone_context():
    # # HACK for CI
    # if os.environ.get("CI"):
    #     return None
    return mgl.create_standalone_context()


@pytest.fixture
def prog_render_depth_pass(standalone_context):
    return standalone_context.program(
        vertex_shader='''
                #version 330

                in vec3 in_vert;

                void main() {
                    gl_Position = vec4(in_vert, 1.0);
                }
            ''',
        fragment_shader='''
                #version 330

                void main() {
                }
            ''',
    )


@pytest.fixture
def prog_draw_depth(standalone_context):
    prog = standalone_context.program(
        vertex_shader='''
                    #version 330

                    in vec2 in_vert;
                    out vec2 uv;

                    void main() {
                        uv = in_vert;
                        gl_Position = vec4(in_vert * 2.0 - 1.0, 0.0, 1.0);
                    }
                ''',
        # from `texture.py`
        fragment_shader='''
                    #version 330

                    uniform sampler2D depth;

                    in vec2 uv;
                    out float fragColor;

                    void main() {
                        float raw_depth_nonlinear = texture(depth, uv).r;
                        fragColor = raw_depth_nonlinear;
                    }
                ''',
    )
    prog['depth'].value = 0
    return prog


@pytest.fixture
def vbo_triangle(standalone_context):
    vertices = np.array([[-1, 1, 0.], [-1, -1, 0.], [1, -1, 0.], ])
    return standalone_context.buffer(vertices.astype('f4').tobytes())


@pytest.fixture
def vbo_quad(standalone_context):
    canvas_fs_quad = np.array(
        [0.0, 0.0, 0.0, 1.0, 1.0, 0.0, 1.0, 1.0]).astype('f4')
    return standalone_context.buffer(canvas_fs_quad.tobytes())


@pytest.fixture
def prog_shadow_test(standalone_context):
    prog = standalone_context.program(
        vertex_shader='''
                #version 330

                in vec2 in_vert;
                out vec2 uv;

                void main() {
                    uv = in_vert;
                    gl_Position = vec4(in_vert * 2.0 - 1.0, 0.0, 1.0);
                }
        ''',
        fragment_shader='''
                #version 330

                uniform sampler2DShadow depth;

                uniform vec2 u_shadow_coord_offset;

                uniform float u_depth_bias;
                uniform float u_depth_triangle;

                in vec2 uv;
                out float fragColor;

                void main() {
                    // Compute depth caster
                    float depth_shadow_caster = u_depth_triangle + u_depth_bias;
                    // Compute shadow coordinates + depth of caster
                    vec3 shadow_coord = vec3(uv, depth_shadow_caster);
                    shadow_coord += vec3(u_shadow_coord_offset, 0.0);
                    // Compute visibility:
                    // -> shadow test: receiver (potential) vs caster
                    float visibility = texture(depth, shadow_coord);
                    // Output the visibility
                    fragColor = visibility;
                }
            ''',
    )
    #
    prog['u_depth_bias'].value = EPSILON_IEEE_754
    depth_triangle_in_obj_space = 0
    depth_triangle_in_view_space = (depth_triangle_in_obj_space + 1) * 0.5
    prog['u_depth_triangle'].value = depth_triangle_in_view_space
    #
    return prog


@pytest.fixture
def fbo_with_rasterised_triangle(standalone_context, vbo_triangle):
    def _build_fbo_with_rasterised_triangle(prog, size=(4, 4), depth_clear=1.0):
        tex_depth = standalone_context.depth_texture(size)  # implicit -> dtype='f4', components=1
        fbo_depth = standalone_context.framebuffer(depth_attachment=tex_depth)
        fbo_depth.use()
        fbo_depth.clear(depth=depth_clear)

        # vertex array object of triangle with depth pass prog
        vao_triangle = standalone_context.simple_vertex_array(prog, vbo_triangle, 'in_vert')
        # Now we render a triangle in there
        vao_triangle.render()

        return fbo_depth, tex_depth

    return _build_fbo_with_rasterised_triangle


@pytest.fixture
def np_triangle_rasterised():
    def _build_np_triangle_rastised(size):
        # It should have 0.5's where the triangle lies.
        depth_value_from_triangle_vertices = 0.0
        # Map [-1, +1] -> [0, 1]
        depth_value_in_depthbuffer = (depth_value_from_triangle_vertices + 1) * 0.5
        return np.array(
            [
                [depth_value_in_depthbuffer] * (size[0] - (j + 1)) + [1.0] * (j + 1)
                for j in range(size[1])
            ]
        )

    return _build_np_triangle_rastised



'''
    Framebuffer scissoring.
    Renders a fullscreen quad four times with different scissor
    values filling each quadrant of the screen with a different color.

    We swap between rendering geometry and using clear.
'''

import numpy as np

import moderngl
from ported._example import Example


class Scissor(Example):
    gl_version = (3, 3)
    title = "Scissor"
    resizable = True
    aspect_ratio = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                in vec2 in_vert;

                void main() {
                    gl_Position = vec4(in_vert, 0.0, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330

                out vec4 fragColor;
                uniform vec4 color;

                void main() {
                    fragColor = color;
                }
            ''',
        )
        quad = [
            -1.0,  1.0,
            -1.0, -1.0,
            1.0, 1.0,
            1.0, -1.0,
        ]
        vbo = self.ctx.buffer(np.array(quad, dtype='f4'))
        self.vao = self.ctx.simple_vertex_array(self.prog, vbo, 'in_vert')

    def render(self, time: float, frame_time: float):
        """Swap between rendering geometry and using clear"""
        fb_width_half = self.wnd.buffer_width // 2
        fb_height_half = self.wnd.buffer_height // 2

        if self.wnd.frames % 2 == 0:
            # upper left (red)
            self.ctx.scissor = 0, fb_height_half, fb_width_half, fb_height_half
            self.prog['color'].value = 1.0, 0.0, 0.0, 0.0
            self.vao.render(mode=moderngl.TRIANGLE_STRIP)

            # upper right (green)
            self.ctx.scissor = fb_width_half, fb_height_half, fb_width_half, fb_height_half
            self.prog['color'].value = 0.0, 1.0, 0.0, 0.0
            self.vao.render(mode=moderngl.TRIANGLE_STRIP)

            # Lower left
            self.ctx.scissor = 0, 0, fb_width_half, fb_height_half
            self.prog['color'].value = 0.0, 0.0, 1.0, 0.0
            self.vao.render(mode=moderngl.TRIANGLE_STRIP)

            # Lower right
            self.ctx.scissor = fb_width_half, 0, fb_width_half, fb_height_half
            self.prog['color'].value = 1.0, 1.0, 1.0, 1.0
            self.vao.render(mode=moderngl.TRIANGLE_STRIP)
        else:
            # upper left (red)
            self.ctx.scissor = 0, fb_height_half, fb_width_half, fb_height_half
            self.ctx.clear(1.0, 0.0, 0.0, 0.0)

            # upper right (green)
            self.ctx.scissor = fb_width_half, fb_height_half, fb_width_half, fb_height_half
            self.ctx.clear(0.0, 1.0, 0.0, 0.0)

            # Lower left
            self.ctx.scissor = 0, 0, fb_width_half, fb_height_half
            self.ctx.clear(0.0, 0.0, 1.0, 0.0)

            # Lower right
            self.ctx.scissor = fb_width_half, 0, fb_width_half, fb_height_half
            self.ctx.clear(1.0, 1.0, 1.0, 1.0)

        # Reset scissor
        self.ctx.scissor = None


if __name__ == '__main__':
    Scissor.run()



from typing import Tuple

from _moderngl import InvalidObject

__all__ = ['Sampler']


class Sampler:
    """
    A Sampler Object is an OpenGL Object that stores the sampling parameters for a Texture access inside of a shader.

    When a sampler object is bound to a texture image unit,
    the internal sampling parameters for a texture bound to the same image unit are all ignored.
    Instead, the sampling parameters are taken from this sampler object.

    Unlike textures, a samplers state can also be changed freely be at any time
    without the sampler object being bound/in use.

    Samplers are bound to a texture unit and not a texture itself. Be careful with leaving
    samplers bound to texture units as it can cause texture incompleteness issues
    (the texture bind is ignored).

    Sampler bindings do clear automatically between every frame so a texture unit
    need at least one bind/use per frame.
    """

    __slots__ = ['mglo', '_glo', 'ctx', 'extra', 'texture']

    def __init__(self):
        self.mglo = None  #: Internal representation for debug purposes only.
        self._glo = None
        self.ctx = None  #: The context this object belongs to
        self.extra = None  #: Any - Attribute for storing user defined objects
        self.texture = None
        raise TypeError()

    def __hash__(self) -> int:
        return id(self)

    def __repr__(self):
        if hasattr(self, '_glo'):
            return f"<{self.__class__.__name__}: {self._glo}>"
        else:
            return f"<{self.__class__.__name__}: INCOMPLETE>"

    def __del__(self):
        if not hasattr(self, "ctx"):
            return

        if self.ctx.gc_mode == "auto":
            self.release()
        elif self.ctx.gc_mode == "context_gc":
            self.ctx.objects.append(self.mglo)

    def use(self, location: int = 0) -> None:
        """
        Bind the sampler to a texture unit.

        Args:
            location (int): The texture unit
        """
        if self.texture is not None:
            self.texture.use(location)
        self.mglo.use(location)

    def clear(self, location: int = 0) -> None:
        """
        Clear the sampler binding on a texture unit.

        Args:
            location (int): The texture unit
        """
        self.mglo.clear(location)

    def release(self) -> None:
        """Release/destroy the ModernGL object."""
        if self.mglo is not None:
            self.mglo.release()
            self.mglo = InvalidObject()

    @property
    def repeat_x(self) -> bool:
        """
        bool: The x repeat flag for the sampler (Default ``True``).

        Example::

            # Enable texture repeat (GL_REPEAT)
            sampler.repeat_x = True

            # Disable texture repeat (GL_CLAMP_TO_EDGE)
            sampler.repeat_x = False
        """
        return self.mglo.repeat_x

    @repeat_x.setter
    def repeat_x(self, value: bool) -> None:
        self.mglo.repeat_x = value

    @property
    def repeat_y(self) -> bool:
        """
        bool: The y repeat flag for the sampler (Default ``True``).

        Example::

            # Enable texture repeat (GL_REPEAT)
            sampler.repeat_y = True

            # Disable texture repeat (GL_CLAMP_TO_EDGE)
            sampler.repeat_y = False
        """
        return self.mglo.repeat_y

    @repeat_y.setter
    def repeat_y(self, value: bool) -> None:
        self.mglo.repeat_y = value

    @property
    def repeat_z(self) -> bool:
        """
        bool: The z repeat flag for the sampler (Default ``True``).

        Example::

            # Enable texture repeat (GL_REPEAT)
            sampler.repeat_z = True

            # Disable texture repeat (GL_CLAMP_TO_EDGE)
            sampler.repeat_z = False
        """
        return self.mglo.repeat_z

    @repeat_z.setter
    def repeat_z(self, value: bool) -> None:
        self.mglo.repeat_z = value

    @property
    def filter(self) -> Tuple[int, int]:
        """
        tuple: The minification and magnification filter for the sampler.

        (Default ``(moderngl.LINEAR. moderngl.LINEAR)``)

        Example::

            sampler.filter == (moderngl.NEAREST, moderngl.NEAREST)
            sampler.filter == (moderngl.LINEAR_MIPMAP_LINEAR, moderngl.LINEAR)
            sampler.filter == (moderngl.NEAREST_MIPMAP_LINEAR, moderngl.NEAREST)
            sampler.filter == (moderngl.LINEAR_MIPMAP_NEAREST, moderngl.NEAREST)
        """
        return self.mglo.filter

    @filter.setter
    def filter(self, value: Tuple[int, int]) -> None:
        self.mglo.filter = value

    @property
    def compare_func(self) -> str:
        """
        tuple: The compare function for a depth textures (Default ``'?'``).

        By default samplers don't have depth comparison mode enabled.
        This means that depth texture values can be read as a ``sampler2D``
        using ``texture()`` in a GLSL shader by default.

        When setting this property to a valid compare mode, ``GL_TEXTURE_COMPARE_MODE``
        is set to ``GL_COMPARE_REF_TO_TEXTURE`` so that texture lookup
        functions in GLSL will return a depth comparison result instead
        of the actual depth value.

        Accepted compare functions::

            .compare_func = ''    # Disale depth comparison completely
            sampler.compare_func = '<='  # GL_LEQUAL
            sampler.compare_func = '<'   # GL_LESS
            sampler.compare_func = '>='  # GL_GEQUAL
            sampler.compare_func = '>'   # GL_GREATER
            sampler.compare_func = '=='  # GL_EQUAL
            sampler.compare_func = '!='  # GL_NOTEQUAL
            sampler.compare_func = '0'   # GL_NEVER
            sampler.compare_func = '1'   # GL_ALWAYS
        """
        return self.mglo.compare_func

    @compare_func.setter
    def compare_func(self, value: str) -> None:
        self.mglo.compare_func = value

    @property
    def anisotropy(self) -> float:
        """
        float: Number of samples for anisotropic filtering (Default ``1.0``).

        The value will be clamped in range ``1.0`` and ``ctx.max_anisotropy``.

        Any value greater than 1.0 counts as a use of anisotropic filtering::

            # Disable anisotropic filtering
            sampler.anisotropy = 1.0

            # Enable anisotropic filtering suggesting 16 samples as a maximum
            sampler.anisotropy = 16.0
        """
        return self.mglo.anisotropy

    @anisotropy.setter
    def anisotropy(self, value: float) -> None:
        self.mglo.anisotropy = value

    @property
    def border_color(self) -> Tuple[float, float, float, float]:
        """
        tuple: The (r, g, b, a) color for the texture border (Default ``(0.0, 0.0, 0.0, 0.0)``).

        When setting this value the ``repeat_`` values are overridden setting the texture wrap to return
        the border color when outside [0, 1] range.

        Example::

            # Red border color
            sampler.border_color = (1.0, 0.0, 0.0, 0.0)
        """
        return self.mglo.border_color

    @border_color.setter
    def border_color(self, value: Tuple[float, float, float, float]) -> None:
        self.mglo.border_color = value

    @property
    def min_lod(self) -> float:
        """
        float: Minimum level-of-detail parameter (Default ``-1000.0``).

        This floating-point value limits the selection of highest resolution mipmap (lowest mipmap level)
        """
        return self.mglo.min_lod

    @min_lod.setter
    def min_lod(self, value: float) -> None:
        self.mglo.min_lod = value

    @property
    def max_lod(self) -> float:
        """
        float: Minimum level-of-detail parameter (Default ``1000.0``).

        This floating-point value limits the selection of the lowest resolution mipmap (highest mipmap level)
        """
        return self.mglo.max_lod

    @max_lod.setter
    def max_lod(self, value: float) -> None:
        self.mglo.max_lod = value

    def assign(self, index: int) -> Tuple["Sampler", int]:
        """
        Helper method for assigning samplers to scopes.

        Example::

            s1 = ctx.sampler(...)
            s2 = ctx.sampler(...)
            ctx.scope(samplers=(s1.assign(0), s1.assign(1)), ...)

        Returns:
            (self, index) tuple
        """
        return (self, index)



"""
Example headless EGL rendering on linux in a VM
"""
import numpy as np
from PIL import Image
import moderngl

ctx = moderngl.create_context(standalone=True, backend='egl')
fbo = ctx.simple_framebuffer((512, 512), components=4)
fbo.use()

vertices = np.array([
    -1.0,  -1.0,   1.0, 0.0, 0.0,
     1.0,  -1.0,   0.0, 1.0, 0.0,
     0.0,   1.0,   0.0, 0.0, 1.0],
    dtype='f4',
)

prog = ctx.program(vertex_shader="""
#version 330
in vec2 in_vert;
in vec3 in_color;
out vec3 color;
void main() {
    gl_Position = vec4(in_vert, 0.0, 1.0);
    color = in_color;
}
""",
fragment_shader="""
#version 330
out vec4 fragColor;
in vec3 color;
void main() {
    fragColor = vec4(color, 1.0);
}
""",
)
vao = ctx.simple_vertex_array(prog, ctx.buffer(vertices), 'in_vert', 'in_color')
vao.render(mode=moderngl.TRIANGLES)

image = Image.frombytes('RGBA', (512, 512), fbo.read(components=4))
image = image.transpose(Image.FLIP_TOP_BOTTOM)
image.save('triangle.png', format='png')


'''
    This examples renders 8 clones of a rectangle with different sizes and positions
    and having the same color
'''

import numpy as np

from ported._example import Example


class InstancedRendering(Example):
    gl_version = (3, 3)
    title = "Instanced Rendering"
    aspect_ratio = 1.0
    window_size = (640, 640)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                in vec2 in_vert;
                in vec2 in_pos;
                in float in_scale;
                in vec3 in_color;

                out vec3 v_color;

                void main() {
                    gl_Position = vec4(in_pos + (in_vert * in_scale), 0.0, 1.0);
                    v_color = in_color;
                }
            ''',
            fragment_shader='''
                #version 330

                in vec3 v_color;
                out vec4 f_color;

                void main() {
                    f_color = vec4(v_color, 1.0);
                }
            ''',
        )

        # Vertex coordinates stored in position_vertex_buffer
        #
        #     B------D
        #     |      |
        #     A------C
        vertices = np.array([
            -0.5, -0.5,
            -0.5, 0.5,
            0.5, -0.5,
            0.5, 0.5,
        ])

        self.position_vertex_buffer = self.ctx.buffer(vertices.astype('f4'))

        # Vertex colors stored in color_buffer
        #
        #     A, B are green
        #     C, D are blue
        colors = np.array([
            0.0, 1.0, 0.0,
            0.0, 1.0, 0.0,
            0.0, 0.0, 1.0,
            0.0, 0.0, 1.0,
        ])
        self.color_buffer = self.ctx.buffer(colors.astype('f4'))

        # (Per instance) positions and scales stored in pos_scale_buffer
        # There are 8 (x_position, y_position, scale) pairs
        position_scale = np.array([
            0.5, 0.0, 0.3,
            0.35, 0.35, 0.2,
            0.0, 0.5, 0.3,
            -0.35, 0.35, 0.2,
            -0.5, 0.0, 0.3,
            -0.35, -0.35, 0.2,
            0.0, -0.5, 0.3,
            0.35, -0.35, 0.2,
        ])
        self.pos_scale_buffer = self.ctx.buffer(position_scale.astype('f4'))

        # Index buffer (also called element buffer)
        # There are 2 trianges to render
        #
        #     A, B, C
        #     B, C, D
        render_indicies = np.array([
            0, 1, 2,
            1, 2, 3
        ])
        self.index_buffer = self.ctx.buffer(render_indicies.astype('i4'))

        # The vao_content is a list of 3-tuples (buffer, format, attribs)
        # the format can have an empty or '/v', '/i', '/r' ending.
        # '/v' attributes are the default
        # '/i` attributes are per instance attributes
        # '/r' attributes are default values for the attributes (per render attributes)
        vao_content = [
            (self.position_vertex_buffer, '2f', 'in_vert'),
            (self.color_buffer, '3f', 'in_color'),
            (self.pos_scale_buffer, '2f 1f/i', 'in_pos', 'in_scale'),
        ]

        self.vao = self.ctx.vertex_array(self.prog, vao_content, self.index_buffer)

    def render(self, time: float, frame_time: float):
        self.ctx.clear(1.0, 1.0, 1.0)
        self.vao.render(instances=8)


if __name__ == '__main__':
    InstancedRendering.run()



import unittest

from common import get_context


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = get_context()

    def test_1(self):
        buf = self.ctx.buffer(b'abc')

        with self.assertRaises(Exception):
            buf.read(4)

        with self.assertRaises(Exception):
            buf.read(offset=-1)

        with self.assertRaises(Exception):
            buf.read(offset=1, size=3)

    def test_2(self):
        buf = self.ctx.buffer(b'123456789')

        with self.assertRaises(Exception):
            buf.read_chunks(1, 4, 1, 6)

        with self.assertRaises(Exception):
            buf.read_chunks(1, 4, -1, 6)

        with self.assertRaises(Exception):
            buf.read_chunks(2, -1, 2, 1)

    def test_3(self):
        buf = self.ctx.buffer(b'123456789')

        with self.assertRaises(Exception):
            buf.read_chunks(2, 2, 3, 3)

        with self.assertRaises(Exception):
            buf.read_chunks(2, -1, -3, 3)

        with self.assertRaises(Exception):
            buf.read_chunks(2, -4, -3, 3)

    def test_4(self):
        buf = self.ctx.buffer(b'123456789')

        with self.assertRaises(Exception):
            buf.read_chunks(3, 0, 2, 2)


if __name__ == '__main__':
    unittest.main()



import numpy as np
from pyrr import Matrix44

import moderngl
from ported._example import Example


def terrain(size):
    vertices = np.dstack(np.mgrid[0:size, 0:size][::-1]) / size
    temp = np.dstack([np.arange(0, size * size - size), np.arange(size, size * size)])
    index = np.pad(temp.reshape(size - 1, 2 * size), [[0, 0], [0, 1]], 'constant', constant_values=-1)
    return vertices, index


class MultiTextireTerrain(Example):
    title = "Multitexture Terrain"
    gl_version = (3, 3)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                uniform mat4 Mvp;
                uniform sampler2D Heightmap;

                in vec2 in_vert;
                out vec2 v_text;

                void main() {
                    vec4 vertex = vec4(in_vert - 0.5, texture(Heightmap, in_vert).r * 0.2, 1.0);
                    gl_Position = Mvp * vertex;
                    v_text = in_vert;
                }
            ''',
            fragment_shader='''
                #version 330

                uniform sampler2D Heightmap;

                uniform sampler2D Color1;
                uniform sampler2D Color2;

                uniform sampler2D Cracks;
                uniform sampler2D Darken;

                in vec2 v_text;

                out vec4 f_color;

                void main() {
                    float height = texture(Heightmap, v_text).r;
                    float border = smoothstep(0.5, 0.7, height);

                    vec3 color1 = texture(Color1, v_text * 7.0).rgb;
                    vec3 color2 = texture(Color2, v_text * 6.0).rgb;

                    vec3 color = color1 * (1.0 - border) + color2 * border;

                    color *= 0.8 + 0.2 * texture(Darken, v_text * 3.0).r;
                    color *= 0.5 + 0.5 * texture(Cracks, v_text * 5.0).r;
                    color *= 0.5 + 0.5 * height;

                    f_color = vec4(color, 1.0);
                }
            ''',
        )

        self.mvp = self.prog['Mvp']

        vertices, index = terrain(32)

        self.vbo = self.ctx.buffer(vertices.astype('f4'))
        self.ibo = self.ctx.buffer(index.astype('i4'))

        vao_content = [
            (self.vbo, '2f', 'in_vert'),
        ]

        self.vao = self.ctx.vertex_array(self.prog, vao_content, self.ibo)

        self.tex0 = self.load_texture_2d('heightmap.jpg')
        self.tex0.build_mipmaps()
        self.tex1 = self.load_texture_2d('grass.jpg')
        self.tex1.build_mipmaps()
        self.tex2 = self.load_texture_2d('rock.jpg')
        self.tex2.build_mipmaps()
        self.tex3 = self.load_texture_2d('cracks.jpg')
        self.tex3.build_mipmaps()
        self.tex4 = self.load_texture_2d('checked.jpg')
        self.tex4.build_mipmaps()

        self.prog['Heightmap'].value = 0
        self.prog['Color1'].value = 1
        self.prog['Color2'].value = 2
        self.prog['Cracks'].value = 3
        self.prog['Darken'].value = 4

    def render(self, time, frame_time):
        angle = time * 0.2
        self.ctx.clear(1.0, 1.0, 1.0)
        self.ctx.enable(moderngl.DEPTH_TEST)

        self.tex0.use(0)
        self.tex1.use(1)
        self.tex2.use(2)
        self.tex3.use(3)
        self.tex4.use(4)

        proj = Matrix44.perspective_projection(45.0, self.aspect_ratio, 0.1, 1000.0)
        lookat = Matrix44.look_at(
            (np.cos(angle), np.sin(angle), 0.8),
            (0.0, 0.0, 0.1),
            (0.0, 0.0, 1.0),
        )

        self.mvp.write((proj * lookat).astype('f4'))
        self.vao.render(moderngl.TRIANGLE_STRIP)


if __name__ == '__main__':
    MultiTextireTerrain.run()



from typing import Any, Tuple

from _moderngl import InvalidObject

__all__ = ['Scope']


class Scope:
    """
    This class represents a Scope object.

    Responsibilities on enter:

    - Set the enable flags.
    - Bind the framebuffer.
    - Assigning textures to texture locations.
    - Assigning buffers to uniform buffers.
    - Assigning buffers to shader storage buffers.

    Responsibilities on exit:

    - Restore the enable flags.
    - Restore the framebuffer.
    """

    __slots__ = [
        'mglo', 'ctx', '_framebuffer', '_textures', '_uniform_buffers',
        '_storage_buffers', '_samplers', 'extra',
    ]

    def __init__(self):
        self.mglo = None  #: Internal representation for debug purposes only.
        self.ctx = None  #: The context this object belongs to
        # Keep references to keep this objects alive
        self._framebuffer = None
        self._textures = None
        self._uniform_buffers = None
        self._storage_buffers = None
        self._samplers = None
        self.extra = None  #: Any - Attribute for storing user defined objects
        raise TypeError()

    def __repr__(self):
        return '<Scope>'

    def __hash__(self) -> int:
        return id(self)

    def __enter__(self):
        self.mglo.begin()
        return self

    def __exit__(self, *args: Tuple[Any]):
        self.mglo.end()

    def __del__(self):
        if not hasattr(self, "ctx"):
            return

        if self.ctx.gc_mode == "auto":
            self.release()
        elif self.ctx.gc_mode == "context_gc":
            self.ctx.objects.append(self.mglo)

    def release(self) -> None:
        """Destroy the Scope object."""
        if self.mglo is not None:
            self._framebuffer = None
            self._textures = None
            self._uniform_buffers = None
            self._storage_buffers = None
            self._samplers = None
            self.mglo.release()
            self.mglo = InvalidObject()



import unittest

import moderngl

from common import get_context


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = get_context()

    def test_buffer_create_bytes(self):
        self.ctx.buffer(b'Hello World!')

    def test_buffer_reserve(self):
        self.ctx.buffer(reserve=1024)

    def test_buffer_reserve_human_readable(self):
        buf = self.ctx.buffer(reserve='2KB')
        self.assertEqual(buf.size, 2 * 1024)

    def test_buffer_data_and_reserve(self):
        with self.assertRaises(moderngl.Error):
            self.ctx.buffer(b'Hello World!', reserve=1024)

    def test_data_size(self):
        buf = self.ctx.buffer(b'\x00\x00\x00\x00')
        self.assertEqual(buf.size, 4)

    def test_reserve_size(self):
        buf = self.ctx.buffer(reserve=8)
        self.assertEqual(buf.size, 8)

    def test_dynamic(self):
        buf = self.ctx.buffer(b'\x00\x00\x00\x00', dynamic=False)
        self.assertFalse(buf.dynamic)

    def test_non_dynamic(self):
        buf = self.ctx.buffer(b'\x00\x00\x00\x00', dynamic=True)
        self.assertTrue(buf.dynamic)


if __name__ == '__main__':
    unittest.main()



import numpy as np

from qtmoderngl import QModernGLWidget
import sys

from PyQt5 import QtWidgets

from renderer_example import HelloWorld2D, PanTool


def vertices():
    x = np.linspace(-1.0, 1.0, 50)
    y = np.random.rand(50) - 0.5
    r = np.ones(50)
    g = np.zeros(50)
    b = np.zeros(50)
    a = np.ones(50)
    return np.dstack([x, y, r, g, b, a])


verts = vertices()
pan_tool = PanTool()


class MyWidget(QModernGLWidget):
    def __init__(self):
        super(MyWidget, self).__init__()
        self.scene = None

    def init(self):
        self.resize(512, 512)
        self.ctx.viewport = (0, 0, 512, 512)
        self.scene = HelloWorld2D(self.ctx)

    def render(self):
        self.screen.use()
        self.scene.clear()
        self.scene.plot(verts)

    def mousePressEvent(self, evt):
        pan_tool.start_drag(evt.x() / 512, evt.y() / 512)
        self.scene.pan(pan_tool.value)
        self.update()

    def mouseMoveEvent(self, evt):
        pan_tool.dragging(evt.x() / 512, evt.y() / 512)
        self.scene.pan(pan_tool.value)
        self.update()

    def mouseReleaseEvent(self, evt):
        pan_tool.stop_drag(evt.x() / 512, evt.y() / 512)
        self.scene.pan(pan_tool.value)
        self.update()


app = QtWidgets.QApplication(sys.argv)
widget = MyWidget()
widget.show()
sys.exit(app.exec_())



import moderngl


class HelloWorld2D:
    def __init__(self, ctx, reserve='4MB'):
        self.ctx = ctx
        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                uniform vec2 Pan;

                in vec2 in_vert;
                in vec4 in_color;

                out vec4 v_color;

                void main() {
                    v_color = in_color;
                    gl_Position = vec4(in_vert - Pan, 0.0, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330

                in vec4 v_color;

                out vec4 f_color;

                void main() {
                    f_color = v_color;
                }
            ''',
        )

        self.vbo = ctx.buffer(reserve='4MB', dynamic=True)
        self.vao = ctx.simple_vertex_array(self.prog, self.vbo, 'in_vert', 'in_color')

    def pan(self, pos):
        self.prog['Pan'].value = pos

    def clear(self, color=(0, 0, 0, 0)):
        self.ctx.clear(*color)

    def plot(self, points, type='line'):
        data = points.astype('f4').tobytes()
        self.vbo.orphan()
        self.vbo.write(data)
        if type == 'line':
            self.ctx.line_width = 1.0
            self.vao.render(moderngl.LINE_STRIP, vertices=len(data) // 24)
        if type == 'points':
            self.ctx.point_size = 3.0
            self.vao.render(moderngl.POINTS, vertices=len(data) // 24)


class PanTool:
    def __init__(self):
        self.total_x = 0.0
        self.total_y = 0.0
        self.start_x = 0.0
        self.start_y = 0.0
        self.delta_x = 0.0
        self.delta_y = 0.0
        self.drag = False

    def start_drag(self, x, y):
        self.start_x = x
        self.start_y = y
        self.drag = True

    def dragging(self, x, y):
        if self.drag:
            self.delta_x = (x - self.start_x) * 2.0
            self.delta_y = (y - self.start_y) * 2.0

    def stop_drag(self, x, y):
        if self.drag:
            self.dragging(x, y)
            self.total_x -= self.delta_x
            self.total_y += self.delta_y
            self.delta_x = 0.0
            self.delta_y = 0.0
            self.drag = False

    @property
    def value(self):
        return (self.total_x - self.delta_x, self.total_y + self.delta_y)



import os

import moderngl_window as mglw


class Example(mglw.WindowConfig):
    gl_version = (3, 3)
    title = "ModernGL Example"
    window_size = (1280, 720)
    aspect_ratio = 16 / 9
    resizable = True

    resource_dir = os.path.normpath(os.path.join(__file__, '../../data'))



"""
Instanced rendering without supplying per instance data
doing offsets with gl_InstanceID in vertex shader.
"""
import numpy as np

import moderngl
from ported._example import Example


class InstancedRendering(Example):
    title = "Instanced Rendering"
    gl_version = (3, 3)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # gl_InstanceID offsets rotation per instance
        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                in vec2 in_vert;
                in vec4 in_color;

                out vec4 v_color;

                uniform float Rotation;
                uniform vec2 Scale;

                void main() {
                    v_color = in_color;
                    float r = Rotation * (0.5 + gl_InstanceID * 0.05);
                    mat2 rot = mat2(cos(r), sin(r), -sin(r), cos(r));
                    gl_Position = vec4((rot * in_vert) * Scale, 0.0, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330

                in vec4 v_color;
                out vec4 f_color;

                void main() {
                    f_color = v_color;
                }
            ''',
        )

        self.scale = self.prog['Scale']
        self.rotation = self.prog['Rotation']

        vertices = np.array([
            # x, y, red, green, blue, alpha
            1.0, 0.0, 1.0, 0.0, 0.0, 0.5,
            -0.5, 0.86, 0.0, 1.0, 0.0, 0.5,
            -0.5, -0.86, 0.0, 0.0, 1.0, 0.5,
        ], dtype='f4')

        self.vbo = self.ctx.buffer(vertices)
        self.vao = self.ctx.vertex_array(
            self.prog,
            [(self.vbo, '2f 4f', 'in_vert', 'in_color')],
        )

    def render(self, time, frame_time):
        self.ctx.clear(1.0, 1.0, 1.0)
        self.ctx.enable(moderngl.BLEND)
        self.scale.value = (0.5, self.aspect_ratio * 0.5)
        self.rotation.value = time
        # For every instanced rendered gl_InstanceID increments by 1
        self.vao.render(instances=10)


if __name__ == '__main__':
    InstancedRendering.run()



from typing import TYPE_CHECKING, Any, List, Optional, Tuple, Union

from _moderngl import InvalidObject

if TYPE_CHECKING:
    from .program import Program
    from .buffer import Buffer

__all__ = ['VertexArray',
           'POINTS', 'LINES', 'LINE_LOOP', 'LINE_STRIP', 'TRIANGLES', 'TRIANGLE_STRIP', 'TRIANGLE_FAN',
           'LINES_ADJACENCY', 'LINE_STRIP_ADJACENCY', 'TRIANGLES_ADJACENCY', 'TRIANGLE_STRIP_ADJACENCY', 'PATCHES']

#: Each vertex represents a point
POINTS = 0x0000
#: Vertices 0 and 1 are considered a line. Vertices 2 and 3 are considered a line.
#: And so on. If the user specifies a non-even number of vertices, then the extra vertex is ignored.
LINES = 0x0001
#: As line strips, except that the first and last vertices are also used as a line.
#: Thus, you get n lines for n input vertices. If the user only specifies 1 vertex,
#: the drawing command is ignored. The line between the first and last vertices happens
#: after all of the previous lines in the sequence.
LINE_LOOP = 0x0002
#: The adjacent vertices are considered lines. Thus, if you pass n vertices, you will get n-1 lines.
#: If the user only specifies 1 vertex, the drawing command is ignored.
LINE_STRIP = 0x0003
#: Vertices 0, 1, and 2 form a triangle. Vertices 3, 4, and 5 form a triangle. And so on.
TRIANGLES = 0x0004
#: Every group of 3 adjacent vertices forms a triangle. The face direction of the
#: strip is determined by the winding of the first triangle. Each successive triangle
#: will have its effective face order reversed, so the system compensates for that
#: by testing it in the opposite way. A vertex stream of n length will generate n-2 triangles.
TRIANGLE_STRIP = 0x0005
#: The first vertex is always held fixed. From there on, every group of 2 adjacent
#: vertices form a triangle with the first. So with a vertex stream, you get a list
#: of triangles like so: (0, 1, 2) (0, 2, 3), (0, 3, 4), etc. A vertex stream of
#: n length will generate n-2 triangles.
TRIANGLE_FAN = 0x0006
#: These are special primitives that are expected to be used specifically with
#: geomtry shaders. These primitives give the geometry shader more vertices
#: to work with for each input primitive. Data needs to be duplicated in buffers.
LINES_ADJACENCY = 0x000A
#: These are special primitives that are expected to be used specifically with
#: geomtry shaders. These primitives give the geometry shader more vertices
#: to work with for each input primitive. Data needs to be duplicated in buffers.
LINE_STRIP_ADJACENCY = 0x000B
#: These are special primitives that are expected to be used specifically with
#: geomtry shaders. These primitives give the geometry shader more vertices
#: to work with for each input primitive. Data needs to be duplicated in buffers.
TRIANGLES_ADJACENCY = 0x000C
#: These are special primitives that are expected to be used specifically with
#: geomtry shaders. These primitives give the geometry shader more vertices
#: to work with for each input primitive. Data needs to be duplicated in buffers.
TRIANGLE_STRIP_ADJACENCY = 0x0000D
#: primitive type can only be used when Tessellation is active. It is a primitive
#: with a user-defined number of vertices, which is then tessellated based on the
#: control and evaluation shaders into regular points, lines, or triangles, depending
#: on the TES's settings.
PATCHES = 0x000E


class VertexArray:
    """
    A VertexArray object is an OpenGL object that stores all of the state needed to supply vertex data.

    It stores the format of the vertex data
    as well as the Buffer objects providing the vertex data arrays.

    In ModernGL, the VertexArray object also stores a reference
    for a :py:class:`Program` object, and some Subroutine information.

    A VertexArray object cannot be instantiated directly, it requires a context.
    Use :py:meth:`Context.vertex_array` or :py:meth:`Context.simple_vertex_array`
    to create one.
    """

    __slots__ = [
        'mglo', '_program', '_index_buffer', '_content', '_index_element_size',
        '_glo', '_mode', 'ctx', 'extra', 'scope'
    ]

    def __init__(self):
        self.mglo = None  #: Internal representation for debug purposes only.
        self._program = None
        self._index_buffer = None
        self._content = None
        self._index_element_size = None
        self._glo = None
        self._mode = None  #: int: The default rendering mode
        self.ctx = None  #: The context this object belongs to
        self.extra = None  #: Any - Attribute for storing user defined objects
        self.scope = None  #: The :py:class:`moderngl.Scope`.
        raise TypeError()

    def __repr__(self) -> str:
        if hasattr(self, 'mglo'):
            return '<VertexArray: %d>' % self.glo
        else:
            return '<VertexArray: INCOMPLETE>'

    def __eq__(self, other: Any) -> bool:
        return type(self) is type(other) and self.mglo is other.mglo

    def __hash__(self) -> int:
        return id(self)

    def __del__(self) -> None:
        if not hasattr(self, "ctx"):
            return

        if self.ctx.gc_mode == "auto":
            self.release()
        elif self.ctx.gc_mode == "context_gc":
            self.ctx.objects.append(self.mglo)

    @property
    def mode(self) -> int:
        """
        int: Get or set the default rendering mode.

        This value is used when ``mode`` is not passed in rendering calls.

        Examples::

            vao.mode = moderngl.TRIANGLE_STRIPS
        """
        return self._mode

    @mode.setter
    def mode(self, value: int) -> None:
        self._mode = value

    @property
    def program(self) -> 'Program':
        """
        Program: The program assigned to the VertexArray.

        The program used when rendering or transforming primitives.
        """
        return self._program

    @property
    def index_buffer(self) -> 'Buffer':
        """Buffer: The index buffer if the index_buffer is set, otherwise ``None``."""
        return self._index_buffer

    @property
    def index_element_size(self) -> int:
        """int: The byte size of each element in the index buffer."""
        return self._index_element_size

    @property
    def vertices(self) -> int:
        """
        int: The number of vertices detected.

        This is the minimum of the number of vertices possible per Buffer.
        The size of the index_buffer determines the number of vertices.
        Per instance vertex attributes does not affect this number.
        """
        return self.mglo.vertices

    @vertices.setter
    def vertices(self, value: int) -> None:
        self.mglo.vertices = int(value)

    @property
    def instances(self) -> int:
        """int: Get or set the number of instances to render."""
        return self.mglo.instances

    @instances.setter
    def instances(self, value: int) -> None:
        self.mglo.instances = int(value)

    @property
    def subroutines(self) -> Tuple[int, ...]:
        """
        tuple: The subroutines assigned to the VertexArray.

        The subroutines used when rendering or transforming primitives.
        """
        return self.mglo.subroutines

    @subroutines.setter
    def subroutines(self, value: Tuple[int, ...]) -> None:
        self.mglo.subroutines = tuple(value)

    @property
    def glo(self) -> int:
        """
        int: The internal OpenGL object.

        This values is provided for debug purposes only.
        """
        return self._glo

    def render(
        self,
        mode: Optional[int] = None,
        vertices: int = -1,
        *,
        first: int = 0,
        instances: int = -1,
    ) -> None:
        """
        The render primitive (mode) must be the same as the input primitive of the GeometryShader.

        Args:
            mode (int): By default :py:data:`TRIANGLES` will be used.
            vertices (int): The number of vertices to transform.

        Keyword Args:
            first (int): The index of the first vertex to start with.
            instances (int): The number of instances.
        """
        if mode is None:
            mode = self._mode

        if self.scope:
            with self.scope:
                self.mglo.render(mode, vertices, first, instances)
        else:
            self.mglo.render(mode, vertices, first, instances)

    def render_indirect(
        self,
        buffer: "Buffer",
        mode: Optional[int] = None,
        count: int = -1,
        *,
        first: int = 0,
    ) -> None:
        """
        The render primitive (mode) must be the same as the input primitive of the GeometryShader.

        The draw commands are 5 integers: (count, instanceCount, firstIndex, baseVertex, baseInstance).

        Args:
            buffer (Buffer): Indirect drawing commands.
            mode (int): By default :py:data:`TRIANGLES` will be used.
            count (int): The number of draws.

        Keyword Args:
            first (int): The index of the first indirect draw command.
        """
        if mode is None:
            mode = self._mode

        if self.scope:
            with self.scope:
                self.mglo.render_indirect(buffer.mglo, mode, count, first)
        else:
            self.mglo.render_indirect(buffer.mglo, mode, count, first)

    def transform(
        self,
        buffer: Union["Buffer", List["Buffer"]],
        mode: Optional[int] = None,
        vertices: int = -1,
        *,
        first: int = 0,
        instances: int = -1,
        buffer_offset: int = 0,
    ) -> None:
        """
        Transform vertices.

        Stores the output in a single buffer.
        The transform primitive (mode) must be the same as
        the input primitive of the GeometryShader.

        Args:
            buffer (Buffer): The buffer to store the output.
            mode (int): By default :py:data:`POINTS` will be used.
            vertices (int): The number of vertices to transform.

        Keyword Args:
            first (int): The index of the first vertex to start with.
            instances (int): The number of instances.
            buffer_offset (int): Byte offset for the output buffer
        """
        if mode is None:
            mode = self._mode

        if isinstance(buffer, (list, tuple)):
            outputs = [buf.mglo for buf in buffer]
        else:
            outputs = [buffer.mglo]

        if self.scope:
            with self.scope:
                self.mglo.transform(outputs, mode, vertices, first, instances, buffer_offset)
        else:
            self.mglo.transform(outputs, mode, vertices, first, instances, buffer_offset)

    def bind(
        self,
        attribute: int,
        cls: str,
        buffer: "Buffer",
        fmt: str,
        *,
        offset: int = 0,
        stride: int = 0,
        divisor: int = 0,
        normalize: bool = False,
    ) -> None:
        """
        Bind individual attributes to buffers.

        Args:
            location (int): The attribute location.
            cls (str): The attribute class. Valid values are ``f``, ``i`` or ``d``.
            buffer (Buffer): The buffer.
            format (str): The buffer format.

        Keyword Args:
            offset (int): The offset.
            stride (int): The stride.
            divisor (int): The divisor.
            normalize (bool): The normalize parameter, if applicable.
        """
        self.mglo.bind(attribute, cls, buffer.mglo, fmt, offset, stride, divisor, normalize)

    def release(self) -> None:
        """Release the ModernGL object."""
        if self.mglo is not None:
            self._program = None
            self._index_buffer = None
            self._content = None
            self.mglo.release()
            self.mglo = InvalidObject()



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import struct

# import GLWindow
# import ModernGL
# from PIL import Image, ImageDraw, ImageFont

# # Window & Context

# wnd = GLWindow.create_window()
# ctx = ModernGL.create_context()

# # Using Pillow to create a Texture

# font = ImageFont.truetype('../data/OpenSans-Regular.ttf', 120)

# msg = 'Hello World!'
# msg_size = font.getsize(msg)

# im = Image.new('RGBA', msg_size, (0, 0, 0, 0))

# draw = ImageDraw.Draw(im)
# draw.text((1, 0), msg, font=font, fill=(0, 0, 0, 255))
# del draw

# tex = ctx.texture(im.size, 4, im.tobytes())
# tex.use()

# # Shaders & Program

# prog = ctx.program(
#     ctx.vertex_shader('''
#         #version 330

#         uniform vec2 Screen;

#         in vec2 vert;
#         in vec2 texcoord;

#         out vec2 v_texcoord;

#         void main() {
#             gl_Position = vec4(vert / Screen * 2.0 - 1.0, 0.0, 1.0);
#             v_texcoord = texcoord;
#         }
#     '''),
#     ctx.fragment_shader('''
#         #version 330

#         uniform sampler2D Texture;

#         in vec2 v_texcoord;
#         out vec4 color;

#         void main() {
#             color = texture(Texture, v_texcoord);
#         }
#     '''),
# ])

# prog.uniforms['Screen'].value = wnd.size

# # Buffer

# vbo = ctx.buffer(struct.pack(
#     '16f',
#     0, 0, 0, 0,
#     0, msg_size[1], 0, -1,
#     msg_size[0], 0, 1, 0,
#     msg_size[0], msg_size[1], 1, -1,
# ))

# # Put everything together

# vao = ctx.simple_vertex_array(prog, vbo, ['vert', 'texcoord'])

# # Main loop

# while wnd.update():
#     ctx.viewport = wnd.viewport
#     ctx.clear(0.9, 0.9, 0.9)
#     ctx.enable(ModernGL.BLEND)
#     vao.render(ModernGL.TRIANGLE_STRIP)



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import os
# import struct

# import GLWindow
# import ModernGL
# from PIL import Image
# from pyrr import Matrix44

# data_folder = os.path.normpath(os.path.join(os.path.dirname(__file__), '..', 'data'))

# wnd = GLWindow.create_window()
# ctx = ModernGL.create_context()

# width, height = wnd.size

# prog = ctx.program(
#     ctx.vertex_shader('''
#         #version 330

#         uniform mat4 Mvp;
#         uniform sampler2D Heightmap;

#         in vec2 in_vert;
#         in vec3 in_color;

#         out vec3 v_color;

#         void main() {
#             v_color = in_color;
#             gl_Position = Mvp * vec4(in_vert.xy * 32.0 - 15.5, texture(Heightmap, in_vert.xy).r * 10.0, 1.0);
#         }
#     '''),
#     ctx.fragment_shader('''
#         #version 330

#         in vec3 v_color;
#         out vec4 f_color;

#         void main() {
#             f_color = vec4(v_color, 1.0);
#         }
#     '''),
# ])

# mvp = prog.uniforms['Mvp']

# vertices = bytearray()
# indices = bytearray()

# for i in range(0, 32):
#     for j in range(0, 32):
#         vertices += struct.pack('2f', i / 32, j / 32)

# for i in range(0, 32 - 1):
#     for j in range(0, 32):
#         indices += struct.pack('2i', i * 32 + j, (i + 1) * 32 + j)
#     indices += struct.pack('i', -1)

# vbo = ctx.buffer(vertices)
# ibo = ctx.buffer(indices)

# vao_content = [
#     (vbo, '2f', ['in_vert']),
# ]

# img = Image.open(os.path.join(data_folder, 'noise.jpg')).convert('L')
# texture = ctx.texture(img.size, 1, img.tobytes())
# texture.use()

# vao = ctx.vertex_array(prog, vao_content, ibo)


# while wnd.update():
#     ctx.viewport = wnd.viewport
#     ctx.clear(0.9, 0.9, 0.9)
#     ctx.enable(ModernGL.DEPTH_TEST)
#     ctx.wireframe = True

#     proj = Matrix44.perspective_projection(45.0, width / height, 0.1, 1000.0)
#     lookat = Matrix44.look_at(
#         (40.0, 30.0, 30.0),
#         (0.0, 0.0, 0.0),
#         (0.0, 0.0, 1.0),
#     )

#     mvp.write((proj * lookat).astype('float32').tobytes())
#     vao.render(ModernGL.TRIANGLE_STRIP)



"""

"""
import numpy as np
import moderngl
import moderngl as mgl
from scipy import ndimage


def test_depth_sampler(standalone_context,
                       prog_render_depth_pass, prog_draw_depth,
                       vbo_triangle, vbo_quad,
                       fbo_with_rasterised_triangle,
                       np_triangle_rasterised):
    ctx = standalone_context
    ctx.enable(mgl.DEPTH_TEST)

    size = (16,) * 2
    fbo_depth, tex_depth = fbo_with_rasterised_triangle(prog_render_depth_pass, size)
    depth_from_dbo = np.frombuffer(tex_depth.read(), dtype=np.dtype('f4')).reshape(size[::-1])

    ############################################################################
    # EXPECTED DATAS                                                                                                   #
    # It should have 0.5's where the triangle lies.                                                                    #
    ############################################################################
    np_triangle_raster = np_triangle_rasterised(size)

    depth_tex_filter = (moderngl.NEAREST, moderngl.NEAREST)
    compare_func = ''

    def _do_test():
        # some stuff should be 1s and 0.5 where the triangle lies.
        np.testing.assert_array_almost_equal(depth_from_dbo, np_triangle_raster)

        fbo_draw_depth = ctx.framebuffer(color_attachments=[ctx.renderbuffer(size, components=1, dtype='f4')])
        fbo_draw_depth.use()

        tex_depth.use(location=0)

        # vertex array object of (fullscreen) quad with render depth prog
        vao_quad = ctx.simple_vertex_array(prog_draw_depth, vbo_quad, 'in_vert')
        vao_quad.render(moderngl.TRIANGLE_STRIP)

        data = np.frombuffer(fbo_draw_depth.read(components=1, dtype='f4'), dtype=np.dtype('f4')).reshape(size[::-1])
        np.testing.assert_array_almost_equal(data, np_triangle_raster)

    def _with_sampler():
        # from: https://github.com/Contraz/demosys-py/blob/master/demosys/opengl/texture.py
        depth_sampler = ctx.sampler(filter=depth_tex_filter, compare_func=compare_func)  # disable depth func
        depth_sampler.use(location=0)
        _do_test()

    def _with_texture_parameters():
        tex_depth.filter = depth_tex_filter
        tex_depth.compare_func = compare_func
        _do_test()

    _with_sampler()
    _with_texture_parameters()


def test_sampler_shadow(standalone_context,
                        prog_render_depth_pass, prog_shadow_test,
                        vbo_triangle, vbo_quad,
                        fbo_with_rasterised_triangle):
    ctx = standalone_context
    ctx.enable(mgl.DEPTH_TEST)

    size = (3,) * 2
    fbo_depth, tex_depth = fbo_with_rasterised_triangle(prog_render_depth_pass, size)

    ############################################################################
    # EXPECTED DATAS                                                                                                   #
    # It should have 1's where the triangle lies.                                                                      #
    ############################################################################
    texel_shadowed = 1
    np_triangle_raster_visibility = np.array(
        [
            [texel_shadowed] * (size[0] - (j + 1)) + [1 - texel_shadowed] * (j + 1)
            for j in range(size[1])
        ]
    )
    ############################################################################

    # No offset - fetch the center pixel (NEAREST)
    prog_shadow_test['u_shadow_coord_offset'].value = (0,) * 2

    depth_tex_filter = (moderngl.NEAREST, moderngl.NEAREST)
    compare_func = '>='

    def _do_test():
        fbo_draw_depth = ctx.framebuffer(color_attachments=[ctx.renderbuffer(size, components=1, dtype='f4')])
        fbo_draw_depth.use()

        tex_depth.use(location=0)

        # https://www.khronos.org/opengl/wiki/Sampler_(GLSL)#Shadow_samplers
        vao_fs_shadow = ctx.simple_vertex_array(prog_shadow_test, vbo_quad, 'in_vert')
        vao_fs_shadow.render(moderngl.TRIANGLE_STRIP)

        data = np.frombuffer(fbo_draw_depth.read(components=1, dtype='f4'), dtype=np.dtype('f4')).reshape(size[::-1])
        np.testing.assert_array_almost_equal(data, np_triangle_raster_visibility)

    def _with_sampler():
        # https://github.com/Contraz/demosys-py/blob/01f285cd3a132012e14a51bf0ae9d7aa5a489b55/demosys/opengl/texture.py#L151
        shadow_sampler = ctx.sampler(filter=depth_tex_filter, compare_func=compare_func)  # enable depth func
        shadow_sampler.use(location=0)
        _do_test()
        shadow_sampler.clear(location=0)

    def _with_texture_paramters():
        tex_depth.filter = depth_tex_filter
        tex_depth.compare_func = compare_func
        _do_test()

    _with_sampler()
    _with_texture_paramters()


def test_sampler_shadow_with_bilinear_interpolation(
        standalone_context,
        prog_render_depth_pass, prog_shadow_test,
        vbo_triangle, vbo_quad,
        fbo_with_rasterised_triangle
):
    ctx = standalone_context
    ctx.enable(mgl.DEPTH_TEST)

    size = (7,) * 2

    fbo_depth, tex_depth = fbo_with_rasterised_triangle(prog_render_depth_pass, size)

    ############################################################################
    # EXPECTED DATAS                                                           #
    ############################################################################
    # Box filter (2x2) on triangle visibility rasterisation
    # is similar to bilinear interpolation with samplerShadow
    # https://docs.scipy.org/doc/scipy/reference/generated/scipy.ndimage.uniform_filter.html#scipy.ndimage.uniform_filter
    np_triangle_raster_visbility = np.array(
        [
            [1.0] * (size[0] - (j + 1)) + [0.0] * (j + 1)
            for j in range(size[1])
        ]
    )
    box_filter_on_triangle_raster_visibility = ndimage.uniform_filter(
        np_triangle_raster_visbility,
        size=2, mode='constant', cval=0, origin=-1)
    ############################################################################

    half_pixel_size = (1.0/size[0]*0.5, 1.0/size[1]*0.5)
    prog_shadow_test['u_shadow_coord_offset'].value = tuple(half_pixel_size)

    def _do_test():
        fbo_draw_depth = ctx.framebuffer(
            color_attachments=[ctx.renderbuffer(size,
                                                components=1, dtype='f4')])
        fbo_draw_depth.use()

        tex_depth.use(location=0)

        # https://www.khronos.org/opengl/wiki/Sampler_(GLSL)#Shadow_samplers
        vao_fs_shadow = ctx.simple_vertex_array(prog_shadow_test, vbo_quad,
                                                'in_vert')
        vao_fs_shadow.render(moderngl.TRIANGLE_STRIP)

        #
        data = np.frombuffer(fbo_draw_depth.read(components=1, dtype='f4'),
                             dtype=np.dtype('f4')).reshape(size[::-1])
        np.testing.assert_array_almost_equal(
            data, box_filter_on_triangle_raster_visibility)

    depth_tex_filter = (moderngl.LINEAR, moderngl.LINEAR)
    compare_func = '>='

    def _with_sampler():
        # https://github.com/Contraz/demosys-py/blob/01f285cd3a132012e14a51bf0ae9d7aa5a489b55/demosys/opengl/texture.py#L151
        shadow_sampler = ctx.sampler(
            filter=depth_tex_filter,
            compare_func=compare_func,  # enable depth func
            # from context.py:
            #   border_color (tuple): The (r, g, b, a) color for the texture border.
            #   When this value is set the ``repeat_`` values are overriden
            #   setting the texture wrap to return the border color when outside ``[0, 1]`` range.
            # border_color=(0, 0, 0, 0),    # seems not to work well ...
            repeat_x=False, repeat_y=False,
        )
        shadow_sampler.use(location=0)
        _do_test()
        shadow_sampler.clear(location=0)

    def _with_texture_paremeters():
        tex_depth.filter = depth_tex_filter
        tex_depth.repeat_x = False
        tex_depth.repeat_y = False
        tex_depth.compare_func = compare_func
        _do_test()

    _with_sampler()
    _with_texture_paremeters()



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import struct

# import GLWindow
# import ModernGL

# # Window & Context

# wnd = GLWindow.create_window()
# ctx = ModernGL.create_context()

# # Shaders & Program

# prog = ctx.program(
#     ctx.vertex_shader('''
#         #version 330

#         in vec2 vert;

#         in vec3 vert_color;
#         out vec3 frag_color;

#         uniform vec2 scale;
#         uniform float rotation;

#         void main() {
#             frag_color = vert_color;
#             mat2 rot = mat2(
#                 cos(rotation), sin(rotation),
#                 -sin(rotation), cos(rotation)
#             );
#             gl_Position = vec4((rot * vert) * scale, 0.0, 1.0);
#         }
#     '''),
#     ctx.fragment_shader('''
#         #version 330

#         in vec3 frag_color;
#         out vec4 color;

#         void main() {
#             color = vec4(frag_color, 1.0);
#         }
#     '''),
# ])

# # Uniforms

# scale = prog.uniforms['scale']
# rotation = prog.uniforms['rotation']

# width, height = wnd.size
# scale.value = (height / width * 0.75, 0.75)

# # Buffer

# vbo = ctx.buffer(struct.pack(
#     '15f',

#     1.0, 0.0,
#     1.0, 0.0, 0.0,

#     -0.5, 0.86,
#     0.0, 1.0, 0.0,

#     -0.5, -0.86,
#     0.0, 0.0, 1.0,
# ))

# # Put everything together

# vao = ctx.simple_vertex_array(prog, vbo, ['vert', 'vert_color'])

# # Main loop

# while wnd.update():
#     ctx.viewport = wnd.viewport
#     ctx.clear(0.9, 0.9, 0.9)
#     rotation.value = wnd.time
#     vao.render()



import unittest

import moderngl

from common import get_context


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = get_context()

    def test_create(self):
        sampler = self.ctx.sampler()
        sampler.use(location=0)
        sampler.clear(location=0)

    def test_defaults(self):
        sampler = self.ctx.sampler()
        self.assertTrue(sampler.anisotropy <= 1.0)
        self.assertTrue(sampler.repeat_x)
        self.assertTrue(sampler.repeat_y)
        self.assertTrue(sampler.repeat_z)
        self.assertEqual(sampler.filter, (moderngl.LINEAR, moderngl.LINEAR))
        self.assertEqual(sampler.compare_func, '?')
        self.assertEqual(sampler.border_color, (0.0, 0.0, 0.0, 0.0))
        self.assertEqual(sampler.min_lod, -1000.0)
        self.assertEqual(sampler.max_lod, 1000.0)

    def test_prop_changes(self):
        sampler = self.ctx.sampler()
        # Change values
        sampler.anisotropy = self.ctx.max_anisotropy
        sampler.filter = (moderngl.NEAREST_MIPMAP_NEAREST, moderngl.NEAREST)
        sampler.compare_func = "<="

        self.assertEqual(sampler.anisotropy, self.ctx.max_anisotropy)
        self.assertEqual(sampler.filter, (moderngl.NEAREST_MIPMAP_NEAREST, moderngl.NEAREST))
        self.assertEqual(sampler.compare_func, "<=")

        # Ensure repeat parameters are set correctly
        sampler.repeat_x = False
        self.assertEqual((sampler.repeat_x, sampler.repeat_y, sampler.repeat_z), (False, True, True))
        sampler.repeat_y = False
        self.assertEqual((sampler.repeat_x, sampler.repeat_y, sampler.repeat_z), (False, False, True))
        sampler.repeat_z = False
        self.assertEqual((sampler.repeat_x, sampler.repeat_y, sampler.repeat_z), (False, False, False))

    def test_border_color(self):
        sampler = self.ctx.sampler()

        # Ensure border color values are set correctly
        colors = [
            (1.0, 0.0, 0.0, 0.0),
            (0.0, 1.0, 0.0, 0.0),
            (0.0, 0.0, 1.0, 0.0),
            (0.0, 0.0, 0.0, 1.0),
        ]
        for color in colors:
            sampler.border_color = color
            self.assertEqual(sampler.border_color, color)

    def test_lod(self):
        sampler = self.ctx.sampler()

        sampler.min_lod = 0.0
        self.assertEqual(sampler.min_lod, 0.0)

        sampler.max_lod = 500.0
        self.assertEqual(sampler.max_lod, 500.0)

    def test_clear_samplers(self):
        self.ctx.clear_samplers(start=0, end=5)
        self.ctx.clear_samplers(start=5, end=10)
        self.ctx.clear_samplers(start=10, end=100)



'''
simple raymarching demo with moderngl

author: minu jeong
'''
import numpy as np

from ported._example import Example


class Raymarching(Example):
    gl_version = (3, 3)
    window_size = (500, 500)
    aspect_ratio = 1.0

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.vaos = []

        program = self.ctx.program(
            vertex_shader=VERTEX_SHADER,
            fragment_shader=FRAGMENT_SHADER
        )

        vertex_data = np.array([
            # x,    y,   z,    u,   v
            -1.0, -1.0, 0.0,  0.0, 0.0,
            +1.0, -1.0, 0.0,  1.0, 0.0,
            -1.0, +1.0, 0.0,  0.0, 1.0,
            +1.0, +1.0, 0.0,  1.0, 1.0,
        ]).astype(np.float32)

        content = [(
            self.ctx.buffer(vertex_data),
            '3f 2f',
            'in_vert', 'in_uv'
        )]

        idx_data = np.array([
            0, 1, 2,
            1, 2, 3
        ]).astype(np.int32)

        idx_buffer = self.ctx.buffer(idx_data)
        self.vao = self.ctx.vertex_array(program, content, idx_buffer)
        self.u_time = program.get("T", 0.0)

    def render(self, time: float, frame_time: float):
        self.u_time.value = time
        self.vao.render()


VERTEX_SHADER = '''
#version 430

in vec3 in_vert;
in vec2 in_uv;
out vec2 v_uv;
void main()
{
    gl_Position = vec4(in_vert.xyz, 1.0);
    v_uv = in_uv;
}
'''

FRAGMENT_SHADER = '''
#version 430

#define FAR 80.0
#define MARCHING_MINSTEP 0
#define MARCHING_STEPS 128
#define MARCHING_CLAMP 0.000001
#define NRM_OFS 0.001
#define AO_OFS 0.01
#define PI 3.141592
#define FOG_DIST 2.5
#define FOG_DENSITY 0.32
#define FOG_COLOR vec3(0.35, 0.37, 0.42)

layout(location=0) uniform float T;

// in vec2 v_uv: screen space coordniate
in vec2 v_uv;

// out color
out vec4 out_color;

// p: sample position
// r: rotation in Euler angles (radian)
vec3 rotate(vec3 p, vec3 r)
{
    vec3 c = cos(r);
    vec3 s = sin(r);
    mat3 rx = mat3(
        1, 0, 0,
        0, c.x, -s.x,
        0, s.x, c.s
    );
    mat3 ry = mat3(
        c.y, 0, s.y,
        0, 1, 0,
        -s.y, 0, c.y
    );
    mat3 rz = mat3(
        c.z, -s.z, 0,
        s.z, c.z, 0,
        0, 0, 1
    );
    return rz * ry * rx * p;
}

// p: sample position
// t: tiling distance
vec3 tile(vec3 p, vec3 t)
{
    return mod(p, t) - 0.5 * t;
}

// p: sample position
// r: radius
float sphere(vec3 p, float r)
{
    return length(p) - r;
}

// p: sample position
// b: width, height, length (scalar along x, y, z axis)
float box(vec3 p, vec3 b)
{
    return length(max(abs(p) - b, 0.0));
}

// c.x, c.y: offset
// c.z: radius
float cylinder(vec3 p, vec3 c)
{
    return length(p.xz - c.xy) - c.z;
}

// a, b: capsule position from - to
// r: radius r
float capsule(vec3 p, vec3 a, vec3 b, float r)
{
    vec3 dp = p - a;
    vec3 db = b - a;
    float h = clamp(dot(dp, db) / dot(db, db), 0.0, 1.0);
    return length(dp - db * h) - r;
}

// p: sample position
// c: cylinder c
// b: box b
float clamp_cylinder(vec3 p, vec3 c, vec3 b)
{
    return max(cylinder(p, c), box(p, b));
}
// a: primitive a
// b: primitive b
// k: blending amount
float blend(float a, float b, float k)
{
    float h = clamp(0.5 + 0.5 * (a - b) / k, 0.0, 1.0);
    return mix(a, b, h) - k * h * (1.0 - h);
}

float displace(vec3 p, float m, float s)
{
    return sin(p.x * m) * sin(p.y * m) * sin(p.z * m) * s;
}

// world
float sample_world(vec3 p, inout vec3 c)
{
    vec3 b_left_pos = p - vec3(-0.8, -0.25, 0.0);
    b_left_pos = rotate(b_left_pos, vec3(T, 0.0, 0.0));
    float d_box_left = box(b_left_pos, vec3(0.4));

    vec3 b_right_pos = p - vec3(+0.8, -0.25, 0.0);
    b_right_pos = rotate(b_right_pos, vec3(0.0, 0.0, T));
    float d_box_right = box(b_right_pos, vec3(0.4));

    vec3 b_up_pos = p - vec3(0.0, 1.05, 0.0);
    b_up_pos = rotate(b_up_pos, vec3(0.0, T, 0.0));
    float d_box_up = box(b_up_pos, vec3(0.4));

    float d_box = FAR;
    d_box = min(d_box, d_box_left);
    d_box = min(d_box, d_box_right);
    d_box = min(d_box, d_box_up);

    vec3 s_pos = p - vec3(0.0, 0.2, 0.0);
    float d_sphere = sphere(s_pos, 0.65);

    float result = blend(d_sphere, d_box, 0.3);

    if (result < FAR)
    {
        c.x = 0.5;
        c.y = 0.75;
        c.z = 0.25;
    }

    return result;
}

// o: origin
// r: ray
// c: color
float raymarch(vec3 o, vec3 r, inout vec3 c)
{
    float t = 0.0;
    vec3 p = vec3(0);
    float d = 0.0;
    for (int i = MARCHING_MINSTEP; i < MARCHING_STEPS; i++)
    {
        p = o + r * t;
        d = sample_world(p, c);
        if (abs(d) < MARCHING_CLAMP)
        {
            return t;
        }
        t += d;
    }
    return FAR;
}

// p: sample surface
vec3 norm(vec3 p)
{
    vec2 o = vec2(NRM_OFS, 0.0);
    vec3 dump_c = vec3(0);
    return normalize(vec3(
        sample_world(p + o.xyy, dump_c) - sample_world(p - o.xyy, dump_c),
        sample_world(p + o.yxy, dump_c) - sample_world(p - o.yxy, dump_c),
        sample_world(p + o.yyx, dump_c) - sample_world(p - o.yyx, dump_c)
    ));
}

void main()
{
    // o: origin
    vec3 o = vec3(0.0, 0.5, -6.0);

    // r: ray
    vec3 r = normalize(vec3(v_uv - vec2(0.5, 0.5), 1.001));

    // l: light
    vec3 l = normalize(vec3(-0.5, -0.2, 0.1));

    // c: albedo
    vec3 c = vec3(0.125);
    float d = raymarch(o, r, c);

    // pixel color
    vec3 color = vec3(0);
    if (d < FAR)
    {
        vec3 p = o + r * d;
        vec3 n = norm(p);

        float lambert = dot(n, l);
        lambert = clamp(lambert, 0.1, 1.0);

        #define SPEC_COLOR vec3(0.85, 0.75, 0.5)
        vec3 h = normalize(o + l);
        float ndh = clamp(dot(n, h), 0.0, 1.0);
        float ndv = clamp(dot(n, -o), 0.0, 1.0);
        float spec = pow((ndh + ndv) + 0.01, 64.0) * 0.25;

        color = c * lambert + SPEC_COLOR * spec;
    }

    // add simple fog
    color = mix(FOG_COLOR, color, clamp(pow(FOG_DIST / abs(d), FOG_DENSITY), 0.0, 1.0));

    out_color = vec4(color, 1.0);
}
'''


if __name__ == '__main__':
    Raymarching.run()



"""ModernGL: High performance rendering for Python 3."""

from .error import *  # noqa
from .buffer import *  # noqa
from .compute_shader import *  # noqa
from .conditional_render import *  # noqa
from .context import *  # noqa
from .framebuffer import *  # noqa
from .program import *  # noqa
from .program_members import *  # noqa
from .query import *  # noqa
from .renderbuffer import *  # noqa
from .scope import *  # noqa
from .texture import *  # noqa
from .texture_3d import *  # noqa
from .texture_array import *  # noqa
from .texture_cube import *  # noqa
from .vertex_array import *  # noqa
from .sampler import *  # noqa

__version__ = '5.7.4'



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import struct
# import sys

# import ModernGL
# from OpenGL.GLUT import (
#     GLUT_DEPTH, GLUT_DOUBLE, GLUT_RGB, glutCreateWindow, glutDisplayFunc, glutInit, glutInitDisplayMode,
#     glutInitWindowSize, glutMainLoop, glutSwapBuffers
# )

# glutInit(sys.argv)
# glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH)
# glutInitWindowSize(800, 600)
# glutCreateWindow(b'Hello World')

# ctx = ModernGL.create_context()

# prog = ctx.program(
#     ctx.vertex_shader('''
#         #version 330

#         in vec2 vert;

#         void main() {
#             gl_Position = vec4(vert, 0.0, 1.0);
#         }
#     '''),
#     ctx.fragment_shader('''
#         #version 330

#         out vec4 color;

#         void main() {
#             color = vec4(0.30, 0.50, 1.00, 1.0);
#         }
#     '''),
# ])

# vbo = ctx.buffer(struct.pack('6f', 0.0, 0.8, -0.6, -0.8, 0.6, -0.8))
# vao = ctx.simple_vertex_array(prog, vbo, ['vert'])


# def display():
#     ctx.clear(0.9, 0.9, 0.9)
#     vao.render()
#     glutSwapBuffers()


# glutDisplayFunc(display)
# glutMainLoop()
